import straxen
import numpy as np
import unittest


class TestCreateVetoIntervals(unittest.TestCase):
    def setUp(self):
        dtype = straxen.plugins.veto_events.veto_event_dtype('nveto_eventbumber')
        dtype += straxen.plugins.veto_events.veto_event_positions_dtype()[2:]
        self.dtype = dtype

        self.events = np.zeros(4, self.dtype)
        self.events['area'] = 1
        self.events['n_hits'] = 1
        self.events['n_contributing_pmt'] = 1
        self.events['time'] = [2, 5, 7, 20]
        self.events['endtime'] = [3, 7, 8, 22]

    def test_empty_inputs(self):
        events = np.zeros(0, self.dtype)
        vetos = straxen.plugins.veto_veto_regions.create_veto_intervals(events, 0, 0, 0, 10, 10)
        assert len(vetos) == 0, 'Empty input must return empty output!'

    def test_concatenate_overlapping_intervals(self):
        left_extension = 1
        right_extension = 4
        vetos = straxen.plugins.veto_veto_regions.create_veto_intervals(self.events,
                                                                        min_area=0,
                                                                        min_hits=0,
                                                                        min_contributing_channels=0,
                                                                        left_extension=left_extension,
                                                                        right_extension=right_extension)
        assert len(vetos) == 2, 'Got the wrong number of veto intervals!'

        time_is_correct = vetos[0]['time'] == self.events['time'][0]-left_extension
        assert time_is_correct, 'First veto event has the wrong time!'
        time_is_correct = vetos[0]['endtime'] == self.events['endtime'][2] + right_extension
        assert time_is_correct, 'First veto event has the wrong endtime!'

        time_is_correct = vetos[1]['time'] == self.events['time'][-1] - left_extension
        assert time_is_correct, 'Second veto event has the wrong time!'
        time_is_correct = vetos[1]['endtime'] == self.events['endtime'][-1] + right_extension
        assert time_is_correct, 'Second veto event has the wrong endtime!'

    def test_thresholds(self):
        events = np.copy(self.events[:1])

        self._test_threshold_type(events, 'area', 'min_area', 2)
        self._test_threshold_type(events, 'n_hits', 'min_hits', 2)
        self._test_threshold_type(events, 'n_contributing_pmt', 'min_contributing_channels', 2)

    @staticmethod
    def _test_threshold_type(events, field, threshold_type, threshold):
        thresholds = {'min_area': 1,
                      'min_hits': 1,
                      'min_contributing_channels': 1}
        thresholds = {key: (threshold if threshold_type == key else 1) for key in thresholds.keys()}

        vetos = straxen.plugins.veto_veto_regions.create_veto_intervals(events,
                                                                        **thresholds,
                                                                        left_extension=0,
                                                                        right_extension=0)
        print(events[field], thresholds, vetos)
        assert len(vetos) == 0, f'Vetos for {threshold_type} threshold should be empty since it is below threshold!'

        events[field] = threshold
        vetos = straxen.plugins.veto_veto_regions.create_veto_intervals(events,
                                                                        **thresholds,
                                                                        left_extension=0,
                                                                        right_extension=0)
        assert len(vetos) == 1, f'{threshold_type} threshold did not work, have a wrong number of vetos!'
        events[field] = 1



import numpy as np
import pandas as pd
import socket
import strax
import straxen
import sys
import warnings
import datetime
import pytz
from os import environ as os_environ

export, __all__ = strax.exporter()
from configparser import NoSectionError


@export
def dataframe_to_wiki(df, float_digits=5, title='Awesome table', 
                      force_int=tuple()):
    """Convert a pandas dataframe to a dokuwiki table 
    (which you can copy-paste onto the XENON wiki)
    :param df: dataframe to convert
    :param float_digits: Round float-ing point values to this number of digits.
    :param title: title of the table.
    """
    table = '^ %s ' % title + '^' * (len(df.columns) - 1) + '^\n'
    table += '^ ' + ' ^ '.join(df.columns) + ' ^\n'

    def do_round(x):
        if isinstance(x, float):
            return round(x, float_digits)
        return x
    force_int = np.where(np.in1d(df.columns.values,
                                 strax.to_str_tuple(force_int)))[0]

    for _, row in df.iterrows():
        table += "| " + ' | '.join([
            str(int(x) if i in force_int else do_round(x))
            for i, x in enumerate(row.values.tolist())]) + ' |\n'
    return table


@export
def print_versions(modules=('strax', 'straxen', 'cutax'), return_string=False):
    """
    Print versions of modules installed.

    :param modules: Modules to print, should be str, tuple or list. E.g.
        print_versions(modules=('strax', 'straxen', 'wfsim',
        'cutax', 'pema'))
    :param return_string: optional. Instead of printing the message,
        return a string
    :return: optional, the message that would have been printed
    """
    message = (f'Working on {socket.getfqdn()} with the following '
               f'versions and installation paths:')
    py_version = sys.version.replace(' (', '\t(').replace('\n', '')
    message += f"\npython\tv{py_version}"
    for m in strax.to_str_tuple(modules):
        try:
            # pylint: disable=exec-used
            exec(f'import {m}')
            # pylint: disable=eval-used
            message += f'\n{m}\tv{eval(m).__version__}\t{eval(m).__path__[0]}'
        except (ModuleNotFoundError, ImportError):
            print(f'{m} is not installed')
    if return_string:
        return message
    print(message)


@export
def utilix_is_configured(header='RunDB', section='xent_database') -> bool:
    """
    Check if we have the right connection to
    :return: bool, can we connect to the Mongo database?
    """
    try:
        return (hasattr(straxen.uconfig, 'get') and
                straxen.uconfig.get(header, section) is not None)
    except NoSectionError:
        return False


@export
class TimeWidgets:

    def __init__(self):
        """Creates interactive time widgets which allow to convert a
        human readble date and time into a start and endtime in unix time
        nano-seconds.
        """
        self._created_widgets = False

    def create_widgets(self):
        """
        Creates time and time zone widget for simpler time querying.

        Note:
            Please be aware that the correct format for the time field
            is HH:MM.
        """
        utcend = datetime.datetime.utcnow()
        deltat = datetime.timedelta(minutes=60)
        utcstart = utcend - deltat
        import ipywidgets as widgets

        self._start_widget = self._create_date_and_time_widget(utcstart, 'Start')
        self._end_widget = self._create_date_and_time_widget(utcend, 'End')
        self._time_zone_widget = self._create_time_zone_widget()
        self._created_widgets = True

        return widgets.VBox([widgets.HBox([self._time_zone_widget, widgets.HTML(value="ns:")]),
                             self._start_widget,
                             self._end_widget])

    def get_start_end(self):
        """
        Returns start and end time of the specfied time interval in
        nano-seconds utc unix time.
        """
        if not self._created_widgets:
            raise ValueError('Please run first "create_widgets". ')

        time_zone = self._time_zone_widget.options[self._time_zone_widget.value][0]

        start, start_ns = self._convert_to_datetime(self._start_widget, time_zone)
        start = start.astimezone(tz=pytz.UTC)
        start = start.timestamp()
        start = int(start * 10**9) + start_ns
        end, end_ns = self._convert_to_datetime(self._end_widget, time_zone)
        end = end.astimezone(tz=pytz.UTC)
        end = end.timestamp()
        end = int(end * 10**9) + end_ns

        if start > end:
            warnings.warn('Start time is larger than endtime are you '
                          'sure you wanted this?')
        return start, end

    @staticmethod
    def _create_date_and_time_widget(date_and_time, widget_describtion):
        import ipywidgets as widgets
        from ipywidgets import Layout
        date = datetime.date(date_and_time.year,
                             date_and_time.month,
                             date_and_time.day)
        date = widgets.DatePicker(description=widget_describtion,
                                  value=date,
                                  layout=Layout(width='225px'),
                                  disabled=False)

        time = "{:02d}:{:02d}".format(int(date_and_time.hour),
                                      int(date_and_time.minute))
        time = widgets.Text(value=time,
                            layout=Layout(width='75px'),
                            disabled=False)

        time_ns = widgets.Text(value='0',
                               layout=Layout(width='150px'),
                               disabled=False)

        return widgets.HBox([date, time, time_ns])

    @staticmethod
    def _create_time_zone_widget():
        import ipywidgets as widgets
        _time_zone_widget = widgets.Dropdown(options=[('CET', 0), ('UTC', 1)],
                                             value=0,
                                             description='Time Zone:',
                                             )
        return _time_zone_widget

    @staticmethod
    def _convert_to_datetime(time_widget, time_zone):
        """
        Converts values of widget into a timezone aware datetime object.

        :param time_widget: Widget Box containing a DatePicker and
            two text widget. The first text widget is used to set a day
            time. The second the time in nano-seconds.
        :param time_zone: pytz.timezone allowed string for a timezone.

        :returns: timezone aware datetime object.
        """
        date_and_time = [c.value for c in time_widget.children]
        try:
            hour_and_minutes = datetime.datetime.strptime(date_and_time[1], '%H:%M')
        except ValueError as e:
            raise ValueError('Cannot convert time into datetime object. '
                             f'Expected the following formating HH:MM. {e}')

        time = datetime.datetime.combine(date_and_time[0], datetime.time())
        time = time.replace(hour=hour_and_minutes.hour,
                            minute=hour_and_minutes.minute)
        time_zone = pytz.timezone(time_zone)
        time = time_zone.localize(time)

        time_ns = int(date_and_time[2])

        return time, time_ns


@export
def convert_array_to_df(array: np.ndarray) -> pd.DataFrame:
    """
    Converts the specified array into a DataFrame drops all higher
    dimensional fields during the process.

    :param array: numpy.array to be converted.
    :returns: DataFrame with higher dimensions dropped.
    """
    keys = [key for key in array.dtype.names if array[key].ndim == 1]
    return pd.DataFrame(array[keys])


@export
def _is_on_pytest():
    """Check if we are on a pytest"""
    return 'PYTEST_CURRENT_TEST' in os_environ



import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import warnings
import strax
import straxen
from mpl_toolkits.axes_grid1 import inset_locator
from datetime import datetime
from .records_matrix import DEFAULT_MAX_SAMPLES
from .daq_waveforms import group_by_daq

export, __all__ = strax.exporter()
__all__ += ['plot_wf']


@straxen.mini_analysis()
def plot_waveform(context,
                  deep=False,
                  show_largest=100,
                  figsize=None,
                  max_samples=DEFAULT_MAX_SAMPLES,
                  ignore_max_sample_warning=True,
                  cbar_loc='lower right',
                  lower_panel_height=2,
                  **kwargs):
    """Plot the sum waveform and optionally per-PMT waveforms

    :param deep: If True, show per-PMT waveform matrix under sum waveform.
    If 'raw', use raw_records instead of records to do so.
    :param show_largest: Show only the largest show_largest peaks.
    :param figsize: Matplotlib figure size for the plot

    Additional options for deep = True or raw:
    :param cbar_loc: location of the intensity color bar. Set to None
    to omit it altogether.
    :param lower_panel_height: Height of the lower panel in terms of
    the height of the upper panel.
    """
    if figsize is None:
        figsize = (10, 6 if deep else 4)

    if not deep:
        context.plot_peaks(**kwargs, show_largest=show_largest, figsize=figsize)

    else:
        f, axes = plt.subplots(2, 1,
                               figsize=figsize,
                               gridspec_kw={'height_ratios': [1, lower_panel_height]})

        plt.sca(axes[0])
        context.plot_peaks(**kwargs, show_largest=show_largest,
                           single_figure=False,
                           xaxis=False)

        plt.sca(axes[1])
        context.plot_records_matrix(**kwargs,
                                    cbar_loc=cbar_loc,
                                    max_samples=max_samples,
                                    ignore_max_sample_warning=ignore_max_sample_warning,
                                    raw=deep == 'raw',
                                    single_figure=False)

        straxen.quiet_tight_layout()
        plt.subplots_adjust(hspace=0)


@straxen.mini_analysis(
    requires=('peaks', 'peak_basics'),
    default_time_selection='touching',
    warn_beyond_sec=60)
def plot_peaks(peaks,
               seconds_range,
               t_reference,
               show_largest=100,
               single_figure=True,
               figsize=(10, 4),
               xaxis=True):
    if single_figure:
        plt.figure(figsize=figsize)
    plt.axhline(0, c='k', alpha=0.2)

    peaks = peaks[np.argsort(-peaks['area'])[:show_largest]]
    peaks = strax.sort_by_time(peaks)

    for p in peaks:
        plot_peak(p,
                  t0=t_reference,
                  color={0: 'gray', 1: 'b', 2: 'g'}[p['type']])

    if xaxis == 'since_start':
        seconds_range_xaxis(seconds_range, t0=seconds_range[0])
    elif xaxis:
        seconds_range_xaxis(seconds_range)
        plt.xlim(*seconds_range)
    else:
        plt.xticks([])
        plt.xlim(*seconds_range)
    plt.ylabel("Intensity [PE/ns]")
    if single_figure:
        plt.tight_layout()


@straxen.mini_analysis(
    requires=('peaks', 'peak_basics'),
    default_time_selection='touching',
    warn_beyond_sec=60)
def plot_hit_pattern(peaks,
                     seconds_range,
                     t_reference,
                     axes=None,
                     vmin=None,
                     log_scale=False,
                     label=None,
                     single_figure=False,
                     xenon1t=False,
                     figsize=(10, 4), ):
    if single_figure:
        plt.figure(figsize=figsize)
    if len(peaks) > 1:
        print(f'warning showing total area of {len(peaks)} peaks')
    straxen.plot_pmts(np.sum(peaks['area_per_channel'], axis=0),
                      axes=axes, vmin=vmin, log_scale=log_scale, label=label,
                      xenon1t=xenon1t)


@straxen.mini_analysis()
def plot_records_matrix(context, run_id,
                        seconds_range,
                        cbar_loc='upper right',
                        raw=False,
                        single_figure=True, figsize=(10, 4),
                        group_by=None,
                        max_samples=DEFAULT_MAX_SAMPLES,
                        ignore_max_sample_warning=False,
                        vmin = None,
                        vmax = None,
                        **kwargs):
    if seconds_range is None:
        raise ValueError(
            "You must pass a time selection (e.g. seconds_range) "
            "to plot_records_matrix.")

    if single_figure:
        plt.figure(figsize=figsize)

    f = context.raw_records_matrix if raw else context.records_matrix

    wvm, ts, ys = f(run_id,
                    max_samples=max_samples,
                    ignore_max_sample_warning=ignore_max_sample_warning,
                    **kwargs)
    if group_by is not None:
        ylabs, wvm_mask = group_by_daq(context, run_id, group_by)
        wvm = wvm[:, wvm_mask]
        plt.ylabel(group_by)
    else:
        plt.ylabel('Channel number')

    # extract min and max from kwargs or set defaults
    if vmin is None:
        vmin = min(0.1 * wvm.max(), 1e-2)
    if vmax is None:
        vmax = wvm.max()

    plt.pcolormesh(
        ts, ys, wvm.T,
        norm=matplotlib.colors.LogNorm(
            vmin=vmin,
            vmax=vmax,
        ),
        cmap=plt.cm.inferno)
    plt.xlim(*seconds_range)

    ax = plt.gca()
    if group_by is not None:
        # Do some magic to convert all the labels to an integer that
        # allows for remapping of the y labels to whatever is provided
        # in the "ylabs", otherwise matplotlib shows nchannels different
        # labels in the case of strings.
        # Make a dict that converts the label to an int
        int_labels = {h: i for i, h in enumerate(set(ylabs))}
        mask = np.ones(len(ylabs), dtype=np.bool_)
        # If the label (int) is different wrt. its neighbour, show it
        mask[1:] = np.abs(np.diff([int_labels[y] for y in ylabs])) > 0
        # Only label the selection
        ax.set_yticks(np.arange(len(ylabs))[mask])
        ax.set_yticklabels(ylabs[mask])
    plt.xlabel('Time [s]')

    if cbar_loc is not None:
        # Create a white box to place the color bar in
        # See https://stackoverflow.com/questions/18211967
        bbox = inset_locator.inset_axes(ax,
                                        width="20%", height="22%",
                                        loc=cbar_loc)
        [bbox.spines[k].set_visible(False) for k in bbox.spines]
        bbox.patch.set_facecolor((1, 1, 1, 0.9))
        bbox.set_xticks([])
        bbox.set_yticks([])

        # Create the actual color bar
        cax = inset_locator.inset_axes(bbox,
                                       width='90%', height='20%',
                                       loc='upper center')
        plt.colorbar(cax=cax,
                     label='Intensity [PE/ns]',
                     orientation='horizontal')
        cax.xaxis.set_major_formatter(
            matplotlib.ticker.FormatStrFormatter('%g'))

    plt.sca(ax)

    if single_figure:
        straxen.quiet_tight_layout()


def seconds_range_xaxis(seconds_range, t0=None):
    """Make a pretty time axis given seconds_range"""
    plt.xlim(*seconds_range)
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    xticks = plt.xticks()[0]
    if not len(xticks):
        return

    # Format the labels
    # I am not very proud of this code...
    def chop(x):
        return np.floor(x).astype(np.int)

    if t0 is None:
        xticks_ns = np.round(xticks * int(1e9)).astype(np.int)
    else:
        xticks_ns = np.round((xticks - xticks[0]) * int(1e9)).astype(np.int)
    sec = chop(xticks_ns // int(1e9))
    ms = chop((xticks_ns % int(1e9)) // int(1e6))
    us = chop((xticks_ns % int(1e6)) // int(1e3))
    samples = chop((xticks_ns % int(1e3)) // 10)

    labels = [str(sec[i]) for i in range(len(xticks))]
    print_ns = np.any(samples != samples[0])
    print_us = print_ns | np.any(us != us[0])
    print_ms = print_us | np.any(ms != ms[0])
    if print_ms and t0 is None:
        labels = [l + f'.{ms[i]:03}' for i, l in enumerate(labels)]
        if print_us:
            labels = [l + r' $\bf{' + f'{us[i]:03}' + '}$'
                      for i, l in enumerate(labels)]
            if print_ns:
                labels = [l + f' {samples[i]:02}0' for i, l in enumerate(labels)]
        plt.xticks(ticks=xticks, labels=labels, rotation=90)
    else:
        labels = list(chop((xticks_ns // 10) * 10))
        labels[-1] = ""
        plt.xticks(ticks=xticks, labels=labels, rotation=0)
    if t0 is None:
        plt.xlabel("Time since run start [sec]")
    else:
        plt.xlabel("Time [ns]")


def plot_peak(p, t0=None, center_time=True, **kwargs):
    x, y = time_and_samples(p, t0=t0)
    kwargs.setdefault('linewidth', 1)

    # Plot waveform
    plt.plot(x, y,
             drawstyle='steps-pre',
             **kwargs)
    if 'linewidth' in kwargs:
        del kwargs['linewidth']
    kwargs['alpha'] = kwargs.get('alpha', 1) * 0.2
    plt.fill_between(x, 0, y, step='pre', linewidth=0, **kwargs)

    # Mark extent with thin black line
    plt.plot([x[0], x[-1]], [y.max(), y.max()],
             c='k', alpha=0.3, linewidth=1)

    # Mark center time with thin black line
    if center_time:
        if t0 is None:
            t0 = p['time']
        ct = (p['center_time'] - t0) / int(1e9)
        plt.axvline(ct, c='k', alpha=0.4, linewidth=1, linestyle='--')


def time_and_samples(p, t0=None):
    """Return (x, y) numpy arrays for plotting the waveform data in p
    using 'steps-pre'.
    Where x is the time since t0 in seconds (or another time_scale),
      and y is intensity in PE / ns.
    :param p: Peak or other similar strax data type
    :param t0: Zero of time in ns since unix epoch
    """
    n = p['length']
    if t0 is None:
        t0 = p['time']
    x = ((p['time'] - t0) + np.arange(n + 1) * p['dt']) / int(1e9)
    y = p['data'][:n] / p['dt']
    return x, np.concatenate([[y[0]], y])



import socket
import re
from tqdm import tqdm
import json
from bson import json_util
import os
import glob
import hashlib
import time
from utilix import xent_collection
import strax
import warnings

try:
    from rucio.client.client import Client
    from rucio.client.rseclient import RSEClient
    from rucio.client.replicaclient import ReplicaClient
    from rucio.client.downloadclient import DownloadClient
    from rucio.client.didclient import DIDClient
    from rucio.common.exception import DataIdentifierNotFound

    rucio_client = Client()
    rse_client = RSEClient()
    replica_client = ReplicaClient()
    did_client = DIDClient()
    download_client = DownloadClient()
    RUCIO_AVAILABLE = True
except (ModuleNotFoundError, RuntimeError):
    warnings.warn("No installation of rucio-clients found. Can't use rucio remote backend.")
    RUCIO_AVAILABLE = False

export, __all__ = strax.exporter()
__all__ += ['RUCIO_AVAILABLE']


class TooMuchDataError(Exception):
    pass


class DownloadError(Exception):
    pass


@export
class RucioFrontend(strax.StorageFrontend):
    """
    Uses the rucio client for the data find.
    """
    local_rses = {'UC_DALI_USERDISK': r'.rcc.'}
    local_did_cache = None
    local_rucio_path = None

    def __init__(self,
                 include_remote=False,
                 download_heavy=False,
                 staging_dir='./strax_data',
                 *args, **kwargs):
        """
        :param include_remote: Flag specifying whether or not to allow rucio downloads from remote sites
        :param download_heavy: option to allow downloading of heavy data through RucioRemoteBackend
        :param args: Passed to strax.StorageFrontend
        :param kwargs: Passed to strax.StorageFrontend
        """
        super().__init__(*args, **kwargs)
        self.readonly = True
        self.collection = xent_collection()

        # check if there is a local rse for the host we are running on
        hostname = socket.getfqdn()
        local_rse = None
        for rse, host_regex in self.local_rses.items():
            if re.search(host_regex, hostname):
                if local_rse is not None:
                    raise ValueError(f"The regex {host_regex} matches two RSEs {rse} and {local_rse}. "
                                     f"I'm not sure what to do with that.")
                local_rse = rse

        # if there is no local host and we don't want to include the
        # remote ones, we can't do anything
        if local_rse is None and not include_remote:
            raise RuntimeError(f"Could not find a local RSE for hostname {hostname}, "
                               f"and include_remote is False.")

        self.local_rse = local_rse
        self.include_remote = include_remote

        self.backends = []
        if local_rse:
            # get the rucio prefix for the local rse, and setup strax
            # rucio backend to read from that path
            rucio_prefix = self.get_rse_prefix(local_rse)
            self.backends.append(RucioLocalBackend(rucio_prefix))
            self.local_rucio_path = rucio_prefix

        if include_remote:
            self.backends.append(RucioRemoteBackend(staging_dir, download_heavy=download_heavy))

    def __repr__(self):
        # List the relevant attributes
        attributes = ('include_remote', 'exclude', 'take_only', 'readonly')
        representation = f'{self.__class__.__module__}.{self.__class__.__name__}'
        for attr in attributes:
            if hasattr(self, attr) and getattr(self, attr):
                representation += f', {attr}: {getattr(self, attr)}'
        return representation

    def find_several(self, keys, **kwargs):
        if not len(keys):
            return []

        ret = []
        for key in keys:
            did = key_to_rucio_did(key)
            if self.did_is_local(did):
                ret.append(('RucioLocalBackend', did))
            else:
                ret.append(False)
        return ret

    def _find(self, key: strax.DataKey, write, allow_incomplete, fuzzy_for, fuzzy_for_options):
        did = key_to_rucio_did(key)
        if allow_incomplete or write:
            raise RuntimeError(f'Allow incomplete/writing is not allowed for '
                               f'{self.__class.__name__} since data might not be '
                               f'continuous')
        if self.did_is_local(did):
            return "RucioLocalBackend", did
        elif self.include_remote:
            # only do this part if we include the remote backend
            try:
                # check if the DID exists
                scope, name = did.split(':')
                did_client.get_did(scope, name)
                return "RucioRemoteBackend", did
            except DataIdentifierNotFound:
                pass

        if fuzzy_for or fuzzy_for_options:
            matches_to = self._match_fuzzy(key,
                                           fuzzy_for,
                                           fuzzy_for_options)
            if matches_to:
                return matches_to

        raise strax.DataNotAvailable

    def get_rse_prefix(self, rse):
        try:
            rse_info = rse_client.get_rse(rse)
            prefix = rse_info['protocols'][0]['prefix']
        except NameError as e:
            if self.local_rse == 'UC_DALI_USERDISK':
                # If we cannot load rucio, let's try the default
                prefix = '/dali/lgrandi/rucio/'
            else:
                raise e
        return prefix

    def did_is_local(self, did):
        """
        Determines whether or not a given did is on a local RSE. If
        there is no local RSE, returns False.

        :param did: Rucio DID string
        :return: boolean for whether DID is local or not.
        """
        try:
            md = self._get_backend("RucioLocalBackend").get_metadata(did)
        except (strax.DataNotAvailable, strax.DataCorrupted):
            return False

        return self._all_chunk_stored(md, did)

    def _all_chunk_stored(self, md: dict, did: str) -> bool:
        """
        Check if all the chunks are stored that are claimed in the
        metadata-file
        """
        scope, name = did.split(':')
        for chunk in md.get('chunks', []):
            if chunk.get('filename'):
                _did = f"{scope}:{chunk['filename']}"
                ch_path = rucio_path(self.local_rucio_path, _did)
                if not os.path.exists(ch_path):
                    return False
        return True

    def _match_fuzzy(self,
                     key: strax.DataKey,
                     fuzzy_for: tuple,
                     fuzzy_for_options: tuple,
                     ) -> tuple:
        # fuzzy for local backend
        pattern = os.path.join(self.get_rse_prefix(self.local_rse),
                               f'xnt_{key.run_id}/*/*/{key.data_type}*metadata.json')
        mds = glob.glob(pattern)
        for md in mds:
            md_dict = read_md(md)
            if self._matches(md_dict['lineage'],
                             # Convert lineage dict to json like to compare
                             json.loads(json.dumps(key.lineage, sort_keys=True)),
                             fuzzy_for,
                             fuzzy_for_options):
                fuzzy_lineage_hash = md_dict['lineage_hash']
                did = f'xnt_{key.run_id}:{key.data_type}-{fuzzy_lineage_hash}'
                self.log.warning(f'Was asked for {key} returning {md}')
                if self._all_chunk_stored(md_dict, did):
                    return 'RucioLocalBackend', did


@export
class RucioLocalBackend(strax.FileSytemBackend):
    """Get data from local rucio RSE"""
    def __init__(self, rucio_dir, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rucio_dir = rucio_dir

    def get_metadata(self, did: str, **kwargs):
        scope, name = did.split(':')
        number, dtype, hsh = parse_did(did)
        metadata_json = f'{dtype}-{hsh}-metadata.json'
        metadata_did = f'{scope}:{metadata_json}'

        metadata_path = rucio_path(self.rucio_dir, metadata_did)
        folder = os.path.join('/', *metadata_path.split('/')[:-1])
        if not os.path.exists(folder):
            raise strax.DataNotAvailable(f"No folder for metadata at {metadata_path}")
        if not os.path.exists(metadata_path):
            raise strax.DataCorrupted(f"Folder exists but no metadata at {metadata_path}")

        with open(metadata_path, mode='r') as f:
            return json.loads(f.read())

    def _read_chunk(self, did, chunk_info, dtype, compressor):
        scope, name = did.split(':')
        did = f"{scope}:{chunk_info['filename']}"
        fn = rucio_path(self.rucio_dir, did)
        return strax.load_file(fn, dtype=dtype, compressor=compressor)

    def _saver(self, **kwargs):
        raise NotImplementedError(
            "Cannot save directly into rucio (yet), upload with admix instead")


@export
class RucioRemoteBackend(strax.FileSytemBackend):
    """Get data from remote Rucio RSE"""

    # datatypes we don't want to download since they're too heavy
    heavy_types = ['raw_records', 'raw_records_nv', 'raw_records_he']

    def __init__(self, staging_dir, download_heavy=False, **kwargs):
        """
        :param staging_dir: Path (a string) where to save data. Must be a writable location.
        :param *args: Passed to strax.FileSystemBackend
        :param **kwargs: Passed to strax.FileSystemBackend
        """

        if os.path.exists(staging_dir):
            if not os.access(staging_dir, os.W_OK):
                raise PermissionError(f"You told the rucio backend to download data to {staging_dir}, "
                                      f"but that path is not writable by your user")
        else:
            try:
                os.makedirs(staging_dir)
            except OSError:
                raise PermissionError(f"You told the rucio backend to download data to {staging_dir}, "
                                      f"but that path is not writable by your user")

        super().__init__(**kwargs)
        self.staging_dir = staging_dir
        self.download_heavy = download_heavy

    def get_metadata(self, dset_did, rse='UC_OSG_USERDISK', **kwargs):
        base_dir = os.path.join(self.staging_dir, did_to_dirname(dset_did))

        # define where the metadata will go (or where it already might be)
        number, dtype, hsh = parse_did(dset_did)
        metadata_file = f"{dtype}-{hsh}-metadata.json"
        metadata_path = os.path.join(base_dir, metadata_file)

        # download if it doesn't exist
        if not os.path.exists(metadata_path):
            metadata_did = f'{dset_did}-metadata.json'
            did_dict = dict(did=metadata_did,
                            base_dir=base_dir,
                            no_subdir=True,
                            rse=rse
                            )
            print(f"Downloading {metadata_did}")
            self._download([did_dict])

        # check again
        if not os.path.exists(metadata_path):
            raise FileNotFoundError(f"No metadata found at {metadata_path}")

        with open(metadata_path, mode='r') as f:
            return json.loads(f.read())

    def _read_chunk(self, dset_did, chunk_info, dtype, compressor, rse="UC_OSG_USERDISK"):
        base_dir = os.path.join(self.staging_dir, did_to_dirname(dset_did))
        chunk_file = chunk_info['filename']
        chunk_path = os.path.join(base_dir, chunk_file)
        if not os.path.exists(chunk_path):
            number, datatype, hsh = parse_did(dset_did)
            if datatype in self.heavy_types and not self.download_heavy:
                error_msg = ("For space reasons we don't want to have everyone "
                             "downloading raw data. If you know what you're "
                             "doing, pass download_heavy=True to the Rucio "
                             "frontend. If not, check your context and/or ask "
                             "someone if this raw data is needed locally.")
                raise DownloadError(error_msg)
            scope, name = dset_did.split(':')
            chunk_did = f"{scope}:{chunk_file}"
            print(f"Downloading {chunk_did}")
            did_dict = dict(did=chunk_did,
                            base_dir=base_dir,
                            no_subdir=True,
                            rse=rse,
                            )
            self._download([did_dict])

        # check again
        if not os.path.exists(chunk_path):
            raise FileNotFoundError(f"No chunk file found at {chunk_path}")

        return strax.load_file(chunk_path, dtype=dtype, compressor=compressor)

    def _saver(self, dirname, metadata):
        raise NotImplementedError("Cannot save directly into rucio (yet), upload with admix instead")

    def _download(self, did_dict_list):
        # need to pass a list of dicts
        # let's try 3 times
        success = False
        _try = 1
        while _try <= 3 and not success:
            if _try > 1:
                for did_dict in did_dict_list:
                    did_dict['rse'] = None
            try:
                download_client.download_dids(did_dict_list)
                success = True
            except KeyboardInterrupt:
                raise
            except Exception:
                sleep = 3**_try
                print(f"Download try #{_try} failed. Sleeping for {sleep} seconds and trying again...")
                time.sleep(sleep)
            _try += 1
        if not success:
            raise DownloadError(f"Error downloading from rucio.")


class RucioSaver(strax.Saver):
    """
    TODO Saves data to rucio if you are the production user
    """

    def __init__(self):
        raise NotImplementedError


def rucio_path(root_dir, did):
    """Convert target to path according to rucio convention.
    See the __hash method here: https://github.com/rucio/rucio/blob/1.20.15/lib/rucio/rse/protocols/protocol.py"""
    scope, filename = did.split(':')
    # disable bandit
    rucio_md5 = hashlib.md5(did.encode('utf-8')).hexdigest() # nosec
    t1 = rucio_md5[0:2]
    t2 = rucio_md5[2:4]
    return os.path.join(root_dir, scope, t1, t2, filename)


def parse_did(did):
    """Parses a Rucio DID and returns a tuple of (number:int, dtype:str, hash: str)"""
    scope, name = did.split(':')
    number = int(scope.split('_')[1])
    dtype, hsh = name.split('-')
    return number, dtype, hsh


def did_to_dirname(did):
    """Takes a Rucio dataset DID and returns a dirname like used by strax.FileSystemBackend"""
    # make sure it's a DATASET did, not e.g. a FILE
    if len(did.split('-')) != 2:
        raise RuntimeError(f"The DID {did} does not seem to be a dataset DID. Is it possible you passed a file DID?")
    dirname = did.replace(':', '-').replace('xnt_', '')
    return dirname


def key_to_rucio_did(key: strax.DataKey) -> str:
    """Convert a strax.datakey to a rucio did field in rundoc"""
    return f'xnt_{key.run_id}:{key.data_type}-{key.lineage_hash}'


def key_to_rucio_meta(key: strax.DataKey) -> str:
    return f'{str(key.data_type)}-{key.lineage_hash}-metadata.json'


def read_md(path: str) -> json:
    with open(path, mode='r') as f:
        md = json.loads(f.read(),
                        object_hook=json_util.object_hook)
    return md


def list_datasets(scope):
    datasets = [d for d in rucio_client.list_dids(scope, {'type': 'dataset'}, type='dataset')]
    return datasets



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@straxen.mini_analysis(requires=('event_basics',))
def load_corrected_positions(context, run_id, events,
                             cmt_version=None,
                             posrec_algos=('mlp', 'gcn', 'cnn')):

    """
    Returns the corrected position for each position algorithm available,
        without the need to reprocess event_basics, as the needed
        information is already stored in event_basics.
    
    :param cmt_version: CMT version to use (it can be a list of same
        length as posrec_algos, if different versions are required for
        different posrec algorithms, default 'local_ONLINE')
    :param posrec_algos: list of position reconstruction algorithms to
        use (default ['mlp', 'gcn', 'cnn'])
    """
    
    posrec_algos = strax.to_str_tuple(posrec_algos)
    
    if cmt_version is None:
        fdc_config = None
        try:
            fdc_config = context.get_single_plugin(run_id, 'event_positions').config['fdc_map']
            cmt_version = fdc_config[1]
        except IndexError as e:
            raise ValueError(f'CMT is not set? Your fdc config is {fdc_config}') from e
    
    if hasattr(cmt_version, '__len__') and not isinstance(cmt_version, str) and len(cmt_version) != len(posrec_algos):
        raise TypeError(f"cmt_version is a list but does not match the posrec_algos ({posrec_algos}) length.")

    cmt_version = (cmt_version, ) * len(posrec_algos) if isinstance(cmt_version, str) else cmt_version

    # Get drift from CMT
    drift_conf = context.get_single_plugin(run_id, 'event_positions').config.get('electron_drift_velocity')
    drift_speed = straxen.get_correction_from_cmt(run_id, drift_conf)
    dtype = []
    
    for algo in posrec_algos:
        for xyzr in 'x y z r'.split():
            dtype += [
                ((f'Interaction {xyzr}-position, field-distortion corrected (cm) - '
                  f'{algo.upper()} posrec algorithm', f'{xyzr}_{algo}'),
                 np.float32),
                     ]
        dtype += [
            ((f'Interaction r-position using observed S2 positions directly (cm) -'
              f' {algo.upper()} posrec algorithm', f'r_naive_{algo}'),
             np.float32),
            ((f'Correction added to r_naive for field distortion (cm) - '
              f'{algo.upper()} posrec algorithm',
              f'r_field_distortion_correction_{algo}'), np.float32),
            ((f'Interaction angular position (radians) - {algo.upper()} '
              f'posrec algorithm', f'theta_{algo}'),
             np.float32)]
        
    dtype += [(('Interaction z-position using mean drift velocity only (cm)', 'z_naive'), np.float32)]
    result = np.zeros(len(events), dtype=dtype)

    z_obs = - drift_speed * events['drift_time']
    
    for algo, v_cmt in zip(posrec_algos, cmt_version):
        fdc_tmp = (f'fdc_map_{algo}', v_cmt, True)
        map_tmp = straxen.get_correction_from_cmt(run_id, fdc_tmp)
        itp_tmp = straxen.InterpolatingMap(straxen.common.get_resource(map_tmp, fmt='binary'))
        itp_tmp.scale_coordinates([1., 1., -drift_speed])

        orig_pos = np.vstack([events[f's2_x_{algo}'], events[f's2_y_{algo}'], z_obs]).T
        r_obs = np.linalg.norm(orig_pos[:, :2], axis=1)
        delta_r = itp_tmp(orig_pos)

        # apply radial correction
        with np.errstate(invalid='ignore', divide='ignore'):
            r_cor = r_obs + delta_r
            scale = r_cor / r_obs

        with np.errstate(invalid='ignore'):
            z_cor = -(z_obs ** 2 - delta_r ** 2) ** 0.5
            invalid = np.abs(z_obs) < np.abs(delta_r)
        z_cor[invalid] = z_obs[invalid]

        result[f'x_{algo}'] = orig_pos[:, 0] * scale
        result[f'y_{algo}'] = orig_pos[:, 1] * scale
        result[f'r_{algo}'] = r_cor
        result[f'r_naive_{algo}'] = r_obs
        result[f'r_field_distortion_correction_{algo}'] = delta_r
        result[f'theta_{algo}'] = np.arctan2(orig_pos[:, 1], orig_pos[:, 0])
        result[f'z_{algo}'] = z_cor

    result['z_naive'] = z_obs
    return result



"""For all of the context, do a quick check to see that we are able to search
a field (i.e. can build the dependencies in the context correctly)
See issue #233 and PR #236"""
from straxen.contexts import xenon1t_dali, xenon1t_led, xenon1t_simulation, fake_daq, demo
from straxen.contexts import xenonnt_led, xenonnt_online, xenonnt_simulation, xenonnt
import straxen
import tempfile
import os


def import_wfsim():
    """Check if we can test the wfsim-related contexts. This is not the case
    for e.g. travis checks as we don't install wfsim by default."""
    try:
        import wfsim
        return True
    except ImportError:
        # We cannot test the wfsim as it is not installed.
        return False

##
# XENONnT
##


def test_xenonnt_online():
    st = xenonnt_online(_database_init=False, use_rucio=False)
    st.search_field('time')


def test_xennonnt():
    if straxen.utilix_is_configured():
        st = xenonnt(_database_init=False, use_rucio=False)
        st.search_field('time')


def test_xennonnt_latest(cmt_version='latest'):
    if straxen.utilix_is_configured():
        st = xenonnt(cmt_version, _database_init=False, use_rucio=False)
        st.search_field('time')


def test_xenonnt_led():
    st = xenonnt_led(_database_init=False, use_rucio=False)
    st.search_field('time')


def test_nt_is_nt_online():
    if not straxen.utilix_is_configured():
        # Cannot contact CMT without the database
        return
    # Test that nT and nT online are the same
    st_online = xenonnt_online(_database_init=False, use_rucio=False)

    st = xenonnt(_database_init=False, use_rucio=False)
    for plugin in st._plugin_class_registry.keys():
        print(f'Checking {plugin}')
        nt_key = st.key_for('0', plugin)
        nt_online_key = st_online.key_for('0', plugin)
        assert str(nt_key) == str(nt_online_key)


def test_xenonnt_simulation():
    if import_wfsim():
        st = xenonnt_simulation()
        st.search_field('time')

##
# XENON1T
##


def test_xenon1t_dali():
    st = xenon1t_dali()
    st.search_field('time')


def test_demo():
    """
    Test the demo context. Since we download the folder to the current
        working directory, make sure we are in a tempfolder where we
        can write the data to
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            print("Temporary directory is ", temp_dir)
            os.chdir(temp_dir)
            st = demo()
            st.search_field('time')
        # On windows, you cannot delete the current process'
        # working directory, so we have to chdir out first.
        finally:
            os.chdir('..')


def test_fake_daq():
    st = fake_daq()
    st.search_field('time')


def test_xenon1t_led():
    st = xenon1t_led()
    st.search_field('time')


def test_xenon1t_simulation():
    if import_wfsim():
        st = xenon1t_simulation()
        st.search_field('time')



import strax
import straxen
import numpy as np

import unittest


class TestRemoveSwtichedOffChannels(unittest.TestCase):

    def setUp(self):
        self.channel_range = (10, 19)
        self.to_pe = np.zeros(20)
        self.to_pe[10:17] = 1

    def test_empty_inputs(self):
        hits = np.zeros(0, strax.hit_dtype)
        hits = straxen.veto_hitlets.remove_switched_off_channels(hits,
                                                                 self.to_pe)
        assert not len(hits), 'Empty input should return an empty result.'

    def test_return(self):
        hits = np.zeros(2, strax.hit_dtype)
        hits[0]['channel'] = 15
        hits[1]['channel'] = 18
        hits_returned = straxen.veto_hitlets.remove_switched_off_channels(hits,
                                                                          self.to_pe)
        assert hits_returned['channel'] == 15, 'Returned a wrong channel.'

        self.to_pe[:] = 1
        hits_returned = straxen.veto_hitlets.remove_switched_off_channels(hits,
                                                                          self.to_pe)
        assert len(hits_returned) == 2, 'Did not return all channels.'



import tempfile
import strax
import numpy as np
from immutabledict import immutabledict
from strax.testutils import recs_per_chunk
import straxen
from straxen.common import pax_file, aux_repo
##
# Tools
##
# Let's make a dummy map for NVeto
nveto_pmt_dummy_df = {'channel': list(range(2000, 2120)),
                      'x': list(range(120)),
                      'y': list(range(120)),
                      'z': list(range(120))}

# Some configs are better obtained from the strax_auxiliary_files repo.
# Let's use small files, we don't want to spend a lot of time downloading
# some file.
testing_config_nT = dict(
    nn_architecture=
    aux_repo + 'f0df03e1f45b5bdd9be364c5caefdaf3c74e044e/fax_files/mlp_model.json',
    nn_weights=
    aux_repo + 'f0df03e1f45b5bdd9be364c5caefdaf3c74e044e/fax_files/mlp_model.h5',
    gain_model=("to_pe_placeholder", True),
    s2_xy_correction_map=pax_file('XENON1T_s2_xy_ly_SR0_24Feb2017.json'),
    elife_conf=('elife_constant', 1e6),
    baseline_samples_nv=10,
    fdc_map=pax_file('XENON1T_FDC_SR0_data_driven_3d_correction_tf_nn_v0.json.gz'),
    gain_model_nv=("adc_nv", True),
    gain_model_mv=("adc_mv", True),
    nveto_pmt_position_map=nveto_pmt_dummy_df,
    s1_xyz_correction_map=pax_file('XENON1T_s1_xyz_lce_true_kr83m_SR0_pax-680_fdc-3d_v0.json'),
    electron_drift_velocity=("electron_drift_velocity_constant", 1e-4),
    s1_aft_map=aux_repo + 'ffdadba3439ae7922b19f5dd6479348b253c09b0/strax_files/s1_aft_UNITY_xyz_XENONnT.json',
    s2_optical_map=aux_repo + '8a6f0c1a4da4f50546918cd15604f505d971a724/strax_files/s2_map_UNITY_xy_XENONnT.json',
    s1_optical_map=aux_repo + '8a6f0c1a4da4f50546918cd15604f505d971a724/strax_files/s1_lce_UNITY_xyz_XENONnT.json',
    electron_drift_time_gate=("electron_drift_time_gate_constant", 2700),
    hit_min_amplitude='pmt_commissioning_initial',
    hit_min_amplitude_nv=20,
    hit_min_amplitude_mv=80,
    hit_min_amplitude_he='pmt_commissioning_initial_he'
)

testing_config_1T = dict(
    hev_gain_model=('1T_to_pe_placeholder', False),
    gain_model=('1T_to_pe_placeholder', False),
    elife_conf=('elife_constant', 1e6),
    electron_drift_velocity=("electron_drift_velocity_constant", 1e-4),
    electron_drift_time_gate=("electron_drift_time_gate_constant", 1700),
)

test_run_id_nT = '008900'
test_run_id_1T = '180423_1021'


@strax.takes_config(
    strax.Option('secret_time_offset', default=0, track=False),
    strax.Option('n_chunks', default=2, track=False,
                 help='Number of chunks for the dummy raw records we are writing here'),
    strax.Option('channel_map', track=False, type=immutabledict,
                 help="frozendict mapping subdetector to (min, max) "
                      "channel number.")
)
class DummyRawRecords(strax.Plugin):
    """
    Provide dummy raw records for the mayor raw_record types
    """
    provides = ('raw_records',
                'raw_records_he',
                'raw_records_nv',
                'raw_records_aqmon',
                'raw_records_aux_mv',
                'raw_records_mv'
                )
    parallel = 'process'
    depends_on = tuple()
    data_kind = immutabledict(zip(provides, provides))
    rechunk_on_save = False
    dtype = {p: strax.raw_record_dtype() for p in provides}

    def setup(self):
        self.channel_map_keys = {'he': 'he',
                                 'nv': 'nveto',
                                 'aqmon': 'aqmon',
                                 'aux_mv': 'aux_mv',
                                 's_mv': 'mv',
                                 }  # s_mv otherwise same as aux in endswith

    def source_finished(self):
        return True

    def is_ready(self, chunk_i):
        return chunk_i < self.config['n_chunks']

    def compute(self, chunk_i):
        t0 = chunk_i + self.config['secret_time_offset']
        if chunk_i < self.config['n_chunks'] - 1:
            # One filled chunk
            r = np.zeros(recs_per_chunk, self.dtype['raw_records'])
            r['time'] = t0
            r['length'] = r['dt'] = 1
            r['channel'] = np.arange(len(r))
        else:
            # One empty chunk
            r = np.zeros(0, self.dtype['raw_records'])

        res = {}
        for p in self.provides:
            rr = np.copy(r)
            # Add detector specific channel offset:
            for key, channel_key in self.channel_map_keys.items():
                if channel_key not in self.config['channel_map']:
                    # Channel map for 1T is different.
                    continue
                if p.endswith(key):
                    s, e = self.config['channel_map'][channel_key]
                    rr['channel'] += s
            res[p] = self.chunk(start=t0, end=t0 + 1, data=rr, data_type=p)
        return res


# Don't concern ourselves with rr_aqmon et cetera
forbidden_plugins = tuple([p for p in
                           straxen.daqreader.DAQReader.provides
                           if p not in DummyRawRecords.provides])


def _run_plugins(st,
                 make_all=False,
                 run_id=test_run_id_nT,
                 **proces_kwargs):
    """
    Try all plugins (except the DAQReader) for a given context (st) to see if
    we can really push some (empty) data from it and don't have any nasty
    problems like that we are referring to some non existant dali folder.
    """

    with tempfile.TemporaryDirectory() as temp_dir:
        st.storage = [strax.DataDirectory(temp_dir)]

        # As we use a temporary directory we should have a clean start
        assert not st.is_stored(run_id, 'raw_records'), 'have RR???'

        # Create event info
        target = 'event_info'
        st.make(run_id=run_id,
                targets=target,
                **proces_kwargs)

        # The stuff should be there
        assert st.is_stored(run_id, target), f'Could not make {target}'

        if not make_all:
            return

        end_targets = set(st._get_end_targets(st._plugin_class_registry))
        for p in end_targets-set(forbidden_plugins):
            st.make(run_id, p)
        # Now make sure we can get some data for all plugins
        all_datatypes = set(st._plugin_class_registry.keys())
        for p in all_datatypes-set(forbidden_plugins):
            should_be_stored = (st._plugin_class_registry[p].save_when ==
                                strax.SaveWhen.ALWAYS)
            if should_be_stored:
                is_stored = st.is_stored(run_id, p)
                assert is_stored, f"{p} did not save correctly!"
    print("Wonderful all plugins work (= at least they don't fail), bye bye")


def _update_context(st, max_workers, fallback_gains=None, nt=True):
    # Change config to allow for testing both multiprocessing and lazy mode
    st.set_context_config({'forbid_creation_of': forbidden_plugins})
    # Ignore strax-internal warnings
    st.set_context_config({'free_options': tuple(st.config.keys())})
    st.register(DummyRawRecords)
    if nt and not straxen.utilix_is_configured():
        st.set_config(testing_config_nT)
        del st._plugin_class_registry['peak_positions_mlp']
        del st._plugin_class_registry['peak_positions_cnn']
        del st._plugin_class_registry['peak_positions_gcn']
        st.register(straxen.PeakPositions1T)
        print(f"Using {st._plugin_class_registry['peak_positions']} for posrec tests")
        st.set_config({'gain_model': fallback_gains})

    elif not nt:
        if straxen.utilix_is_configured():
            # Set some placeholder gain as this takes too long for 1T to load from CMT
            st.set_config({k: v for k, v in testing_config_1T.items() if
                           k in ('hev_gain_model', 'gain_model')})
        else:
            st.set_config(testing_config_1T)

    if max_workers - 1:
        st.set_context_config({
            'allow_multiprocess': True,
            'allow_lazy': False,
            'timeout': 60,  # we don't want to build travis for ever
        })
    print('--- Plugins ---')
    for k, v in st._plugin_class_registry.items():
        print(k, v)


def _test_child_options(st, run_id):
    """
    Test which checks if child options are handled correctly.
    """
    # Register all used plugins
    plugins = []
    already_seen = []
    for data_type in st._plugin_class_registry.keys():
        if data_type in already_seen or data_type in straxen.DAQReader.provides:
            continue

        p = st.get_single_plugin(run_id, data_type)
        plugins.append(p)
        already_seen += p.provides

    # Loop over all plugins and check if child options were propagated to the parent:
    for p in plugins:
        for option_name, option in p.takes_config.items():
            # Check if option is a child option:
            if option.child_option:
                # Get corresponding parent option. Do not have to test if
                # parent option name is defined this is already done in strax
                parent_name = option.parent_option_name

                # Now check if parent config was replaced with child:
                t = p.config[parent_name] == p.config[option_name]
                assert t, (f'This is strange the child option "{option_name}" was set to '
                           f'{p.config[option_name]}, but the corresponding parent config'
                           f' "{parent_name}" has the value {p.config[parent_name]}. '
                           f'Please check the options of {p.__class__.__name__} and if '
                           'it is a child plugin (child_plugin=True)!')

                # Test if parent names were removed from the lineage:
                t = parent_name in p.lineage[p.provides[-1]][2]
                assert not t, (f'Found "{parent_name}" in the lineage of {p.__class__.__name__}. '
                               f'This should not have happend since "{parent_name}" is a child of '
                               f'"{option_name}"!')


##
# Tests
##


def test_1T(ncores=1):
    if ncores == 1:
        print('-- 1T lazy mode --')
    st = straxen.contexts.xenon1t_dali()
    _update_context(st, ncores, nt=False)

    # Register the 1T plugins for this test as well
    st.register_all(straxen.plugins.x1t_cuts)
    for _plugin, _plugin_class in st._plugin_class_registry.items():
        if 'cut' in str(_plugin).lower():
            _plugin_class.save_when = strax.SaveWhen.ALWAYS

    # Run the test
    _run_plugins(st, make_all=True, max_wokers=ncores, run_id=test_run_id_1T)

    # Test issue #233
    st.search_field('cs1')

    # set all the configs to be non-CMT
    st.set_config(testing_config_1T)
    _test_child_options(st, test_run_id_1T)

    print(st.context_config)


def test_nT(ncores=1):
    if ncores == 1:
        print('-- nT lazy mode --')
    st = straxen.contexts.xenonnt_online(_database_init=straxen.utilix_is_configured(),
                                         use_rucio=False)
    offline_gain_model = ("to_pe_placeholder", True)
    _update_context(st, ncores, fallback_gains=offline_gain_model, nt=True)
    # Lets take an abandoned run where we actually have gains for in the CMT
    _run_plugins(st, make_all=True, max_wokers=ncores, run_id=test_run_id_nT)
    # Test issue #233
    st.search_field('cs1')
    # Test of child plugins:
    _test_child_options(st, test_run_id_nT)
    print(st.context_config)


def test_nT_mutlticore():
    print('nT multicore')
    test_nT(2)

# Disable the test below as it saves some time in travis and gives limited new
# information as most development is on nT-plugins.
# def test_1T_mutlticore():
#     print('1T multicore')
#     test_1T(2)



import strax
import straxen

import numpy as np
import unittest


class TestMergeIntervals(unittest.TestCase):

    def setUp(self):
        self.intervals = np.zeros(4, dtype=strax.time_fields)
        self.intervals['time'] = [2, 3, 7, 20]
        self.intervals['endtime'] = [4, 7, 8, 22]

    def test_empty_intervals(self):
        intervals = np.zeros(0, dtype=strax.time_fields)
        intervals = straxen.merge_intervals(intervals)
        assert len(intervals) == 0, 'Empty input should return empty intervals!'

    def test_merge_overlapping_intervals(self):
        intervals = straxen.merge_intervals(self.intervals)

        assert len(intervals) == 2, 'Got the wrong number of intervals!'

        time_is_correct = intervals[0]['time'] == self.intervals['time'][0]
        assert time_is_correct, 'First interval has the wrong time!'
        time_is_correct = intervals[0]['endtime'] == self.intervals['endtime'][-2]
        assert time_is_correct, 'First interval has the wrong endtime!'

        time_is_correct = intervals[-1]['time'] == self.intervals['time'][-1]
        assert time_is_correct, 'Second interval has the wrong time!'
        time_is_correct = intervals[-1]['endtime'] == self.intervals['endtime'][-1]
        assert time_is_correct, 'Second interval has the wrong endtime!'


class TestCoincidence(unittest.TestCase):

    def setUp(self):
        self.intervals = np.zeros(8, dtype=strax.time_fields)
        self.intervals['time'] = [3, 6, 9, 12, 15, 18, 21, 38]
        self.intervals['endtime'] = [5, 8, 10, 13, 16, 19, 23, 42]

    def test_empty_inputs(self):
        intervals = np.zeros(0, dtype=strax.time_fields)
        intervals = straxen.find_coincidence(intervals)
        assert len(intervals) == 0, 'Empty input should return empty intervals!'

    def test_without_coincidence(self):
        resolving_time = 10
        truth_time = np.array([self.intervals['time'][0],
                               self.intervals['time'][-1]])
        truth_endtime = np.array([self.intervals['time'][-2],
                                  self.intervals['time'][-1]]) + resolving_time
        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=1,
                               pre_trigger=0,
                               n_concidences_truth=2,
                               times_truth=truth_time,
                               endtime_truth=truth_endtime
                               )
        pre_trigger = 2
        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=1,
                               pre_trigger=pre_trigger,
                               n_concidences_truth=2,
                               times_truth=truth_time - pre_trigger,
                               endtime_truth=truth_endtime
                               )

    def test_even_fold(self):
        resolving_time = 10
        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=2,
                               pre_trigger=0,
                               n_concidences_truth=1,
                               times_truth=self.intervals['time'][0],
                               endtime_truth=self.intervals['time'][-3] + resolving_time,
                               )
        pre_trigger = 2
        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=2,
                               pre_trigger=pre_trigger,
                               n_concidences_truth=1,
                               times_truth=self.intervals['time'][0] - pre_trigger,
                               endtime_truth=self.intervals['time'][-3] + resolving_time,
                               )

        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=4,
                               pre_trigger=pre_trigger,
                               n_concidences_truth=1,
                               times_truth=self.intervals['time'][0] - pre_trigger,
                               endtime_truth=self.intervals['time'][-5] + resolving_time,
                               )

    def test_odd_fold(self):
        resolving_time = 10
        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=3,
                               pre_trigger=0,
                               n_concidences_truth=1,
                               times_truth=self.intervals['time'][0],
                               endtime_truth=self.intervals['time'][-4] + resolving_time,
                               )
        pre_trigger = 2
        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=3,
                               pre_trigger=pre_trigger,
                               n_concidences_truth=1,
                               times_truth=self.intervals['time'][0] - pre_trigger,
                               endtime_truth=self.intervals['time'][-4] + resolving_time,
                               )

        self._test_coincidence(resolving_time=resolving_time,
                               coincidence=5,
                               pre_trigger=pre_trigger,
                               n_concidences_truth=0,
                               times_truth=self.intervals['time'][:0],
                               endtime_truth=self.intervals['time'][:0],
                               )

    def _test_coincidence(self, resolving_time, coincidence, pre_trigger,
                          n_concidences_truth, times_truth, endtime_truth):
        coincidence = straxen.find_coincidence(self.intervals,
                                               nfold=coincidence,
                                               resolving_time=resolving_time,
                                               pre_trigger=pre_trigger)
        number_coincidence_correct = len(coincidence) == n_concidences_truth
        assert number_coincidence_correct, 'Have not found the correct number of coincidences'

        time_is_correct = np.all(coincidence['time'] == times_truth)
        assert time_is_correct, 'Coincidence does not have the correct time'

        endtime_is_correct = np.all(coincidence['endtime'] == endtime_truth)
        print(coincidence['endtime'], endtime_truth)
        assert endtime_is_correct, 'Coincidence does not have the correct endtime'



from immutabledict import immutabledict
import strax
import straxen
from copy import deepcopy

common_opts = dict(
    register_all=[
        straxen.event_processing,
        straxen.double_scatter,
    ],
    # Register all peak/pulse processing by hand as 1T does not need to have
    # the high-energy plugins.
    register=[
        straxen.PulseProcessing,
        straxen.Peaklets,
        straxen.PeakletClassification,
        straxen.MergedS2s,
        straxen.Peaks,
        straxen.PeakBasics,
        straxen.PeakProximity],
    check_available=('raw_records', 'peak_basics'),
    store_run_fields=(
        'name', 'number',
        'start', 'end', 'livetime', 'mode'))

xnt_common_config = dict(
    n_tpc_pmts=straxen.n_tpc_pmts,
    n_top_pmts=straxen.n_top_pmts,
    gain_model=("to_pe_model", "ONLINE", True),
    gain_model_nv=("to_pe_model_nv", "ONLINE", True),
    gain_model_mv=("to_pe_model_mv", "ONLINE", True),
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        # Channels must be listed in a ascending order!
        tpc=(0, 493),
        he=(500, 752),  # high energy
        aqmon=(790, 807),
        aqmon_nv=(808, 815),  # nveto acquisition monitor
        tpc_blank=(999, 999),
        mv=(1000, 1083),
        aux_mv=(1084, 1087),  # Aux mv channel 2 empty  1 pulser  and 1 GPS
        mv_blank=(1999, 1999),
        nveto=(2000, 2119),
        nveto_blank=(2999, 2999)),
    # Clustering/classification parameters
    # Event level parameters
    s2_xy_correction_map=('s2_xy_map', "ONLINE", True),
    fdc_map=('fdc_map', "ONLINE", True),
    s1_xyz_correction_map=("s1_xyz_map", "ONLINE", True),
    g1=0.1426,
    g2=11.55,
)
# these are placeholders to avoid calling cmt with non integer run_ids. Better solution pending.
# s1,s2 and fd corrections are still problematic
xnt_simulation_config = deepcopy(xnt_common_config)
xnt_simulation_config.update(gain_model=("to_pe_placeholder", True),
                             gain_model_nv=("adc_nv", True),
                             gain_model_mv=("adc_mv", True),
                             elife_conf=('elife_constant', 1e6),
                             )

# Plugins in these files have nT plugins, E.g. in pulse&peak(let)
# processing there are plugins for High Energy plugins. Therefore do not
# st.register_all in 1T contexts.
xnt_common_opts = common_opts.copy()
xnt_common_opts.update({
    'register': common_opts['register'] + [straxen.PeakPositionsCNN,
                                           straxen.PeakPositionsMLP,
                                           straxen.PeakPositionsGCN,
                                           straxen.PeakPositionsNT,
                                           straxen.PeakBasicsHighEnergy,
                                           straxen.PeaksHighEnergy,
                                           straxen.PeakletsHighEnergy,
                                           straxen.PeakletClassificationHighEnergy,
                                           straxen.MergedS2sHighEnergy,
                                           straxen.EventInfo,
                                          ],
    'register_all': common_opts['register_all'] + [straxen.veto_veto_regions,
                                                   straxen.nveto_recorder,
                                                   straxen.veto_pulse_processing,
                                                   straxen.veto_hitlets,
                                                   straxen.veto_events,
                                                   straxen.acqmon_processing,
                                                   straxen.pulse_processing,
                                                   straxen.peaklet_processing,
                                                   straxen.online_monitor,
                                                   straxen.event_area_per_channel,
                                                   straxen.event_patternfit,
                                                   ],
    'use_per_run_defaults': False,
})

##
# XENONnT
##


def xenonnt(cmt_version='global_ONLINE', **kwargs):
    """XENONnT context"""
    st = straxen.contexts.xenonnt_online(**kwargs)
    st.apply_cmt_version(cmt_version)
    return st


def xenonnt_online(output_folder='./strax_data',
                   use_rucio=False,
                   we_are_the_daq=False,
                   _minimum_run_number=7157,
                   _maximum_run_number=None,
                   _database_init=True,
                   _forbid_creation_of=None,
                   _rucio_path='/dali/lgrandi/rucio/',
                   _include_rucio_remote=False,
                   _raw_path='/dali/lgrandi/xenonnt/raw',
                   _processed_path='/dali/lgrandi/xenonnt/processed',
                   _add_online_monitor_frontend=False,
                   _context_config_overwrite=None,
                   **kwargs):
    """
    XENONnT online processing and analysis

    :param output_folder: str, Path of the strax.DataDirectory where new
        data can be stored
    :param use_rucio: bool, whether or not to use the rucio frontend
    :param we_are_the_daq: bool, if we have admin access to upload data
    :param _minimum_run_number: int, lowest number to consider
    :param _maximum_run_number: Highest number to consider. When None
        (the default) consider all runs that are higher than the
        minimum_run_number.
    :param _database_init: bool, start the database (for testing)
    :param _forbid_creation_of: str/tuple, of datatypes to prevent form
        being written (raw_records* is always forbidden).
    :param _include_rucio_remote: allow remote downloads in the context
    :param _rucio_path: str, path of rucio
    :param _raw_path: str, common path of the raw-data
    :param _processed_path: str. common path of output data
    :param _context_config_overwrite: dict, overwrite config
    :param _add_online_monitor_frontend: bool, should we add the online
        monitor storage frontend.
    :param kwargs: dict, context options
    :return: strax.Context
    """
    context_options = {
        **straxen.contexts.xnt_common_opts,
        **kwargs}

    st = strax.Context(
        config=straxen.contexts.xnt_common_config,
        **context_options)
    st.register([straxen.DAQReader, straxen.LEDCalibration])

    st.storage = [
        straxen.RunDB(
            readonly=not we_are_the_daq,
            minimum_run_number=_minimum_run_number,
            maximum_run_number=_maximum_run_number,
            runid_field='number',
            new_data_path=output_folder,
            rucio_path=_rucio_path,
        )] if _database_init else []
    if not we_are_the_daq:
        st.storage += [
            strax.DataDirectory(
                _raw_path,
                readonly=True,
                take_only=straxen.DAQReader.provides),
            strax.DataDirectory(
                _processed_path,
                readonly=True,
            )]
        if output_folder:
            st.storage.append(
                strax.DataDirectory(output_folder))

        st.context_config['forbid_creation_of'] = straxen.daqreader.DAQReader.provides
        if _forbid_creation_of is not None:
            st.context_config['forbid_creation_of'] += strax.to_str_tuple(_forbid_creation_of)

    # if we said so, add the rucio frontend to storage
    if use_rucio:
        st.storage.append(straxen.rucio.RucioFrontend(
            include_remote=straxen.RUCIO_AVAILABLE,
            staging_dir=output_folder,
        ))

    # Only the online monitor backend for the DAQ
    if _database_init and (_add_online_monitor_frontend or we_are_the_daq):
        st.storage += [straxen.OnlineMonitor(
            readonly=not we_are_the_daq,
            take_only=('veto_intervals',
                       'online_peak_monitor',
                       'event_basics',))]

    # Remap the data if it is before channel swap (because of wrongly cabled
    # signal cable connectors) These are runs older than run 8797. Runs
    # newer than 8796 are not affected. See:
    # https://github.com/XENONnT/straxen/pull/166 and
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:dsg:daq:sector_swap
    st.set_context_config({'apply_data_function': (straxen.remap_old,
                                                   straxen.check_loading_allowed,
                                                   )})
    if _context_config_overwrite is not None:
        st.set_context_config(_context_config_overwrite)

    return st


def xenonnt_led(**kwargs):
    st = xenonnt_online(**kwargs)
    st.context_config['check_available'] = ('raw_records', 'led_calibration')
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(
        replace=True,
        config=st.config,
        storage=st.storage,
        **st.context_config)
    st.register([straxen.DAQReader, straxen.LEDCalibration])
    return st


def xenonnt_simulation(output_folder='./strax_data'):
    import wfsim
    st = strax.Context(
        storage=strax.DataDirectory(output_folder),
        config=dict(detector='XENONnT',
                    fax_config='fax_config_nt_design.json',
                    check_raw_record_overlaps=True,
                    **straxen.contexts.xnt_simulation_config,
                    ),
        **straxen.contexts.xnt_common_opts)
    st.register(wfsim.RawRecordsFromFaxNT)
    return st

##
# XENON1T
##


x1t_context_config = {
    **common_opts,
    **dict(
        check_available=('raw_records', 'records', 'peaklets',
                         'events', 'event_info'),
        free_options=('channel_map',),
        use_per_run_defaults=True,
        store_run_fields=tuple(
            [x for x in common_opts['store_run_fields'] if x != 'mode']
            + ['trigger.events_built', 'reader.ini.name']))}
x1t_context_config.update(
    dict(register=common_opts['register'] +
                  [straxen.PeakPositions1T,
                   straxen.RecordsFromPax,
                   straxen.EventInfo1T,
         ]))

x1t_common_config = dict(
    check_raw_record_overlaps=False,
    allow_sloppy_chunking=True,
    n_tpc_pmts=248,
    n_top_pmts=127,
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        tpc=(0, 247),
        diagnostic=(248, 253),
        aqmon=(254, 999)),
    # Records
    hev_gain_model=('to_pe_model', "v1", False),
    pmt_pulse_filter=(
        0.012, -0.119,
        2.435, -1.271, 0.357, -0.174, -0., -0.036,
        -0.028, -0.019, -0.025, -0.013, -0.03, -0.039,
        -0.005, -0.019, -0.012, -0.015, -0.029, 0.024,
        -0.007, 0.007, -0.001, 0.005, -0.002, 0.004, -0.002),
    hit_min_amplitude='XENON1T_SR1',
    tail_veto_threshold=int(1e5),
    save_outside_hits=(3, 3),
    # Peaklets
    peaklet_gap_threshold=350,
    gain_model=('to_pe_model', "v1", False),
    peak_split_gof_threshold=(
        None,  # Reserved
        ((0.5, 1), (3.5, 0.25)),
        ((2, 1), (4.5, 0.4))),
    peak_min_pmts=2,
    # MergedS2s
    s2_merge_gap_thresholds=((1.7, 5.0e3), (4.0, 500.), (5.0, 0.)),
    # Peaks
    # Smaller right extension since we applied the filter
    peak_right_extension=30,
    s1_max_rise_time=60,
    s1_max_rise_time_post100=150,
    s1_min_coincidence=3,
    # Events*
    left_event_extension=int(0.3e6),
    right_event_extension=int(1e6),
    elife_conf=('elife_xenon1t', 'v1', False),
    electron_drift_velocity=("electron_drift_velocity_constant", 1.3325e-4),
    max_drift_length=96.9,
    electron_drift_time_gate=("electron_drift_time_gate_constant", 1700),
)


def demo():
    """Return strax context used in the straxen demo notebook"""
    straxen.download_test_data()

    st = strax.Context(
        storage=[strax.DataDirectory('./strax_data'),
                 strax.DataDirectory('./strax_test_data',
                                     deep_scan=True,
                                     provide_run_metadata=True,
                                     readonly=True)],
        forbid_creation_of=straxen.daqreader.DAQReader.provides,
        config=dict(**x1t_common_config),
        **x1t_context_config)

    # Use configs that are always available
    st.set_config(dict(
        hev_gain_model=('1T_to_pe_placeholder', False),
        gain_model=('1T_to_pe_placeholder', False),
        elife_conf=('elife_constant', 1e6),
        electron_drift_velocity=("electron_drift_velocity_constant", 1.3325e-4),
        ))
    return st


def fake_daq():
    """Context for processing fake DAQ data in the current directory"""
    st = strax.Context(
        storage=[strax.DataDirectory('./strax_data'),
                 # Fake DAQ puts run doc JSON in same folder:
                 strax.DataDirectory('./from_fake_daq',
                                     provide_run_metadata=True,
                                     readonly=True)],
        config=dict(daq_input_dir='./from_fake_daq',
                    daq_chunk_duration=int(2e9),
                    daq_compressor='lz4',
                    n_readout_threads=8,
                    daq_overlap_chunk_duration=int(2e8),
                    **x1t_common_config),
        **x1t_context_config)
    st.register(straxen.Fake1TDAQReader)
    return st


def xenon1t_dali(output_folder='./strax_data', build_lowlevel=False, **kwargs):
    context_options = {
        **x1t_context_config,
        **kwargs}

    st = strax.Context(
        storage=[
            strax.DataDirectory(
                '/dali/lgrandi/xenon1t/strax_converted/raw',
                take_only='raw_records',
                provide_run_metadata=True,
                readonly=True),
            strax.DataDirectory(
                '/dali/lgrandi/xenon1t/strax_converted/processed',
                readonly=True),
            strax.DataDirectory(output_folder)],
        config=dict(**x1t_common_config),
        # When asking for runs that don't exist, throw an error rather than
        # starting the pax converter
        forbid_creation_of=(
            straxen.daqreader.DAQReader.provides if build_lowlevel
            else straxen.daqreader.DAQReader.provides + ('records', 'peaklets')),
        **context_options)
    return st


def xenon1t_led(**kwargs):
    st = xenon1t_dali(**kwargs)
    st.context_config['check_available'] = ('raw_records', 'led_calibration')
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(
        replace=True,
        config=st.config,
        storage=st.storage,
        **st.context_config)
    st.register([straxen.RecordsFromPax, straxen.LEDCalibration])
    return st


def xenon1t_simulation(output_folder='./strax_data'):
    import wfsim
    st = strax.Context(
        storage=strax.DataDirectory(output_folder),
        config=dict(
            fax_config='fax_config_1t.json',
            detector='XENON1T',
            **x1t_common_config),
        **x1t_context_config)
    st.register(wfsim.RawRecordsFromFax1T)
    return st



from straxen.misc import TimeWidgets


def test_widgets():
    tw = TimeWidgets()
    wig = tw.create_widgets()
    start, end = tw.get_start_end()

    assert isinstance(start, int) and isinstance(end, int), "Should have returned unix time in ns as integer!"
    assert end > start, "By default end should be larger than start"

    # Now manually change time zone and compare:
    wig.children[0].children[0].value = 1
    start_utc, end_utc = tw.get_start_end()

    h_in_ns_unix = 60*60*10**9
    unix_conversion_worked = start_utc - start == h_in_ns_unix or start_utc - start == 2 * h_in_ns_unix
    assert unix_conversion_worked
    unix_conversion_worked = start_utc - end == h_in_ns_unix or start_utc - end == 2 * h_in_ns_unix
    assert unix_conversion_worked


def test_change_in_fields():
    tw = TimeWidgets()
    wig = tw.create_widgets()
    start, end = tw.get_start_end()

    # Modify the nano-second value:
    wig.children[1].children[2].value = '20'
    wig.children[2].children[2].value = '20'

    start20, end20 = tw.get_start_end()
    assert start20 - start == 20, 'Start nano-second field did not update.'
    assert end20 - end == 20, 'End nano-second field did not update.'

    # Modify Minutes:
    time = wig.children[1].children[1].value
    minutes = int(time[-2:])
    minutes *= 60*10**9
    wig.children[1].children[1].value = time[:-2] + '00'  # .value is a string "HH:MM"

    start00, _ = tw.get_start_end()
    assert start20 - start00 == minutes, 'Time field did not update its value!'



import strax
import straxen
import numpy as np
from scipy.special import betainc, gammaln, loggamma
import warnings
export, __all__ = strax.exporter()


@export
@strax.takes_config(
    strax.Option('s1_optical_map', help='S1 (x, y, z) optical/pattern map.',
                 default='XENONnT_s1_xyz_patterns_LCE_corrected_qes_MCva43fa9b_wires.pkl'),
    strax.Option('s2_optical_map', help='S2 (x, y) optical/pattern map.',
                 default='XENONnT_s2_xy_patterns_LCE_corrected_qes_MCva43fa9b_wires.pkl'),
    strax.Option('s1_aft_map', help='Date drive S1 area fraction top map.',
                 default='s1_aft_dd_xyz_XENONnT_Kr83m_41500eV_25May2021.json'),
    strax.Option('mean_pe_per_photon', help='Mean of full VUV single photon response',
                 default=1.2),
    strax.Option('gain_model',
                 help='PMT gain model. Specify as (model_type, model_config)'),
    strax.Option('n_tpc_pmts', type=int,
                 help='Number of TPC PMTs'),
    strax.Option('n_top_pmts', type=int,
                 help='Number of top TPC PMTs'),
    strax.Option('s1_min_area_pattern_fit',
                 help='Skip EventPatternFit reconstruction if S1 area (PE) is less than this',
                 default=3),
    strax.Option('s2_min_area_pattern_fit',
                 help='Skip EventPatternFit reconstruction if S2 area (PE) is less than this',
                 default=10),
    strax.Option('store_per_channel', default=False, type=bool,
                 help='Store normalized LLH per channel for each peak'),
    strax.Option('max_r_pattern_fit', default=straxen.tpc_r, type=float,
                 help='Maximal radius of the peaks where llh calculation will be performed'),
)
class EventPatternFit(strax.Plugin):
    '''
    Plugin that provides patter information for events
    '''
    
    depends_on = ('event_area_per_channel', 'event_basics', 'event_positions')
    provides = 'event_pattern_fit'
    __version__ = '0.0.7'

    def infer_dtype(self):
        dtype = [('s2_2llh', np.float32,
                  'Modified Poisson likelihood value for main S2 in the event'),
                 ('alt_s2_2llh', np.float32,
                  'Modified Poisson likelihood value for alternative S2'),
                 ('s1_2llh', np.float32,
                  'Modified Poisson likelihood value for main S1'),
                 ('s1_top_2llh', np.float32,
                  'Modified Poisson likelihood value for main S1, calculated from top array'),
                 ('s1_bottom_2llh', np.float32,
                  'Modified Poisson likelihood value for main S1, calculated from bottom array'),
                 ('s1_area_fraction_top_continuous_probability', np.float32,
                  'Continuous binomial test for S1 area fraction top'),
                 ('s1_area_fraction_top_discrete_probability', np.float32,
                  'Discrete binomial test for S1 area fraction top'),
                 ('s1_photon_fraction_top_continuous_probability', np.float32,
                  'Continuous binomial test for S1 photon fraction top'),
                 ('s1_photon_fraction_top_discrete_probability', np.float32,
                  'Discrete binomial test for S1 photon fraction top'),
                 ('alt_s1_area_fraction_top_continuous_probability', np.float32,
                  'Continuous binomial test for alternative S1 area fraction top'),
                 ('alt_s1_area_fraction_top_discrete_probability', np.float32,
                  'Discrete binomial test for alternative S1 area fraction top'),
                 ('alt_s1_photon_fraction_top_continuous_probability', np.float32,
                  'Continuous binomial test for alternative S1 photon fraction top'),
                 ('alt_s1_photon_fraction_top_discrete_probability', np.float32,
                  'Discrete binomial test for alternative S1 photon fraction top')]
        
        if self.config['store_per_channel']:
            dtype += [
                (('2LLH per channel for main S2', 's2_2llh_per_channel'),
                 np.float32, (self.config['n_top_pmts'], )),
                (('2LLH per channel for alternative S2', 'alt_s2_2llh_per_channel'),
                 np.float32, (self.config['n_top_pmts'], )),
                (('Pattern main S2', 's2_pattern'),
                 np.float32, (self.config['n_top_pmts'], )),
                (('Pattern alt S2', 'alt_s2_pattern'),
                 np.float32, (self.config['n_top_pmts'], )),
                (('Pattern for main S1', 's1_pattern'),
                 np.float32, (self.config['n_tpc_pmts'], )),
                (('2LLH per channel for main S1', 's1_2llh_per_channel'),
                 np.float32, (self.config['n_tpc_pmts'], )),
            ]
        dtype += strax.time_fields
        return dtype
    
    def setup(self):
        self.mean_pe_photon = self.config['mean_pe_per_photon']
        
        # Getting S1 AFT maps
        self.s1_aft_map = straxen.InterpolatingMap(
            straxen.get_resource(
                self.config['s1_aft_map'],
                fmt=self._infer_map_format(self.config['s1_aft_map'])))
                    
        # Getting optical maps
        self.s1_pattern_map = straxen.InterpolatingMap(
            straxen.get_resource(
                self.config['s1_optical_map'],
                fmt=self._infer_map_format(self.config['s1_optical_map'])))
        self.s2_pattern_map = straxen.InterpolatingMap(
            straxen.get_resource(
                self.config['s2_optical_map'],
                fmt=self._infer_map_format(self.config['s2_optical_map'])))
        
        # Getting gain model to get dead PMTs
        self.to_pe = straxen.get_correction_from_cmt(self.run_id, self.config['gain_model'])
        self.dead_PMTs = np.where(self.to_pe == 0)[0]
        self.pmtbool = ~np.in1d(np.arange(0, self.config['n_tpc_pmts']), self.dead_PMTs)
        self.pmtbool_top = self.pmtbool[:self.config['n_top_pmts']]
        self.pmtbool_bottom = self.pmtbool[self.config['n_top_pmts']:self.config['n_tpc_pmts']]
        
    def compute(self, events):
        
        result = np.zeros(len(events), dtype=self.dtype)
        result['time'] = events['time']
        result['endtime'] = strax.endtime(events)

        # Computing LLH values for S1s
        self.compute_s1_llhvalue(events, result)
        
        # Computing LLH values for S2s
        self.compute_s2_llhvalue(events, result)
        
        # Computing binomial test for s1 area fraction top
        s1_area_fraction_top_probability = np.vectorize(_s1_area_fraction_top_probability)
        positions = np.vstack([events['x'], events['y'], events['z']]).T
        aft_prob = self.s1_aft_map(positions)
        
        # main s1 events
        mask_s1 = ~np.isnan(aft_prob)
        mask_s1 &= ~np.isnan(events['s1_area'])
        mask_s1 &= ~np.isnan(events['s1_area_fraction_top'])
        
        # default value is nan, it will be overwrite if the event satisfy the requirements
        result['s1_area_fraction_top_continuous_probability'][:] = np.nan
        result['s1_area_fraction_top_discrete_probability'][:] = np.nan
        result['s1_photon_fraction_top_continuous_probability'][:] = np.nan
        result['s1_photon_fraction_top_discrete_probability'][:] = np.nan
        
        # compute binomial test only if we have events that have valid aft prob, s1 area and s1 aft
        if np.sum(mask_s1):
            arg = aft_prob[mask_s1], events['s1_area'][mask_s1], events['s1_area_fraction_top'][mask_s1]
            result['s1_area_fraction_top_continuous_probability'][mask_s1] = s1_area_fraction_top_probability(*arg)
            result['s1_area_fraction_top_discrete_probability'][mask_s1] = s1_area_fraction_top_probability(*arg, 'discrete')
            arg = aft_prob[mask_s1], events['s1_area'][mask_s1]/self.config['mean_pe_per_photon'], events['s1_area_fraction_top'][mask_s1]
            result['s1_photon_fraction_top_continuous_probability'][mask_s1] = s1_area_fraction_top_probability(*arg)
            result['s1_photon_fraction_top_discrete_probability'][mask_s1] = s1_area_fraction_top_probability(*arg, 'discrete')
        
        # alternative s1 events
        mask_alt_s1 = ~np.isnan(aft_prob)
        mask_alt_s1 &= ~np.isnan(events['alt_s1_area'])
        mask_alt_s1 &= ~np.isnan(events['alt_s1_area_fraction_top'])
        
        # default value is nan, it will be ovewrite if the event satisfy the requirments
        result['alt_s1_area_fraction_top_continuous_probability'][:] = np.nan
        result['alt_s1_area_fraction_top_discrete_probability'][:] = np.nan
        result['alt_s1_photon_fraction_top_continuous_probability'][:] = np.nan
        result['alt_s1_photon_fraction_top_discrete_probability'][:] = np.nan
        
        # compute binomial test only if we have events that have valid aft prob, alt s1 area and alt s1 aft
        if np.sum(mask_alt_s1):
            arg = aft_prob[mask_alt_s1], events['alt_s1_area'][mask_alt_s1], events['alt_s1_area_fraction_top'][mask_alt_s1]
            result['alt_s1_area_fraction_top_continuous_probability'][mask_alt_s1] = s1_area_fraction_top_probability(*arg)
            result['alt_s1_area_fraction_top_discrete_probability'][mask_alt_s1] = s1_area_fraction_top_probability(*arg, 'discrete')
            arg = aft_prob[mask_alt_s1], events['alt_s1_area'][mask_alt_s1]/self.config['mean_pe_per_photon'], events['alt_s1_area_fraction_top'][mask_alt_s1]
            result['alt_s1_photon_fraction_top_continuous_probability'][mask_alt_s1] = s1_area_fraction_top_probability(*arg)
            result['alt_s1_photon_fraction_top_discrete_probability'][mask_alt_s1] = s1_area_fraction_top_probability(*arg, 'discrete')
                
        return result

    def compute_s1_llhvalue(self, events, result):
        # Selecting S1s for pattern fit calculation
        # - must exist (index != -1)
        # - must have total area larger minimal one
        # - must have positive AFT
        x, y, z = events['x'], events['y'], events['z']
        cur_s1_bool = events['s1_area']>self.config['s1_min_area_pattern_fit']
        cur_s1_bool &= events['s1_index']!=-1
        cur_s1_bool &= events['s1_area_fraction_top']>=0
        cur_s1_bool &= np.isfinite(x)
        cur_s1_bool &= np.isfinite(y)
        cur_s1_bool &= np.isfinite(z)
        cur_s1_bool &= (x**2 + y**2) < self.config['max_r_pattern_fit']**2
        
        # default value is nan, it will be ovewrite if the event satisfy the requirments
        result['s1_2llh'][:] = np.nan
        result['s1_top_2llh'][:] = np.nan
        result['s1_bottom_2llh'][:] = np.nan
        
        # Making expectation patterns [ in PE ]
        if np.sum(cur_s1_bool):
            s1_map_effs = self.s1_pattern_map(np.array([x, y, z]).T)[cur_s1_bool, :]
            s1_area = events['s1_area'][cur_s1_bool]
            s1_pattern = s1_area[:, None]*(s1_map_effs[:, self.pmtbool])/np.sum(s1_map_effs[:, self.pmtbool], axis=1)[:, None] 

            s1_pattern_top = (events['s1_area_fraction_top'][cur_s1_bool]*s1_area)
            s1_pattern_top = s1_pattern_top[:, None]*((s1_map_effs[:, :self.config['n_top_pmts']])[:, self.pmtbool_top])
            s1_pattern_top /= np.sum((s1_map_effs[:, :self.config['n_top_pmts']])[:, self.pmtbool_top], axis=1)[:, None] 
            s1_pattern_bottom = ((1-events['s1_area_fraction_top'][cur_s1_bool])*s1_area)
            s1_pattern_bottom = s1_pattern_bottom[:, None]*((s1_map_effs[:, self.config['n_top_pmts']:])[:, self.pmtbool_bottom])
            s1_pattern_bottom /= np.sum((s1_map_effs[:, self.config['n_top_pmts']:])[:, self.pmtbool_bottom], axis=1)[:, None] 

            # Getting pattern from data
            s1_area_per_channel_ = events['s1_area_per_channel'][cur_s1_bool,:]
            s1_area_per_channel = s1_area_per_channel_[:, self.pmtbool]
            s1_area_per_channel_top = (s1_area_per_channel_[:, :self.config['n_top_pmts']])[:, self.pmtbool_top]
            s1_area_per_channel_bottom = (s1_area_per_channel_[:, self.config['n_top_pmts']:])[:, self.pmtbool_bottom]

            # Top and bottom
            arg1 = s1_pattern/self.mean_pe_photon, s1_area_per_channel, self.mean_pe_photon
            arg2 = s1_area_per_channel/self.mean_pe_photon, s1_area_per_channel, self.mean_pe_photon
            norm_llh_val = (neg2llh_modpoisson(*arg1) - neg2llh_modpoisson(*arg2))
            result['s1_2llh'][cur_s1_bool] = np.sum(norm_llh_val, axis=1)

            # If needed to stire - store only top and bottom array, but not together
            if self.config['store_per_channel']:
                # Storring pattern information
                store_patterns = np.zeros((s1_pattern.shape[0], self.config['n_tpc_pmts']) )  
                store_patterns[:, self.pmtbool] = s1_pattern
                result['s1_pattern'][cur_s1_bool] = store_patterns
                # Storing actual LLH values
                store_2LLH_ch = np.zeros((norm_llh_val.shape[0], self.config['n_tpc_pmts']) )
                store_2LLH_ch[:, self.pmtbool] = norm_llh_val
                result['s1_2llh_per_channel'][cur_s1_bool] = store_2LLH_ch

            # Top
            arg1 = s1_pattern_top/self.mean_pe_photon, s1_area_per_channel_top, self.mean_pe_photon
            arg2 = s1_area_per_channel_top/self.mean_pe_photon, s1_area_per_channel_top, self.mean_pe_photon
            norm_llh_val = (neg2llh_modpoisson(*arg1) - neg2llh_modpoisson(*arg2))
            result['s1_top_2llh'][cur_s1_bool] = np.sum(norm_llh_val, axis=1)

            # Bottom
            arg1 = s1_pattern_bottom/self.mean_pe_photon, s1_area_per_channel_bottom, self.mean_pe_photon
            arg2 = s1_area_per_channel_bottom/self.mean_pe_photon, s1_area_per_channel_bottom, self.mean_pe_photon
            norm_llh_val = (neg2llh_modpoisson(*arg1) - neg2llh_modpoisson(*arg2))
            result['s1_bottom_2llh'][cur_s1_bool] = np.sum(norm_llh_val, axis=1)
            
    def compute_s2_llhvalue(self, events, result):
        for t_ in ['s2', 'alt_s2']:
            # Selecting S2s for pattern fit calculation
            # - must exist (index != -1)
            # - must have total area larger minimal one
            # - must have positive AFT
            x, y = events[t_+'_x'], events[t_+'_y']
            cur_s2_bool = (events[t_+'_area']>self.config['s2_min_area_pattern_fit'])
            cur_s2_bool &= (events[t_+'_index']!=-1)
            cur_s2_bool &= (events[t_+'_area_fraction_top']>0)
            cur_s2_bool &= (x**2 + y**2) < self.config['max_r_pattern_fit']**2
            
            # default value is nan, it will be ovewrite if the event satisfy the requirments
            result[t_+'_2llh'][:] = np.nan
            
            # Making expectation patterns [ in PE ]
            if np.sum(cur_s2_bool):
                s2_map_effs = self.s2_pattern_map(np.array([x, y]).T)[cur_s2_bool, 0:self.config['n_top_pmts']]
                s2_map_effs = s2_map_effs[:, self.pmtbool_top]
                s2_top_area = (events[t_+'_area_fraction_top']*events[t_+'_area'])[cur_s2_bool]
                s2_pattern  = s2_top_area[:, None]*s2_map_effs/np.sum(s2_map_effs, axis=1)[:,None]

                # Getting pattern from data
                s2_top_area_per_channel = events[t_+'_area_per_channel'][cur_s2_bool, 0:self.config['n_top_pmts']]
                s2_top_area_per_channel = s2_top_area_per_channel[:, self.pmtbool_top]

                # Calculating LLH, this is shifted Poisson
                # we get area expectation and we need to scale them to get
                # photon expectation
                norm_llh_val = (neg2llh_modpoisson(
                                     mu    = s2_pattern/self.mean_pe_photon, 
                                     areas = s2_top_area_per_channel, 
                                     mean_pe_photon=self.mean_pe_photon)
                                        - 
                                neg2llh_modpoisson(
                                     mu    = s2_top_area_per_channel/self.mean_pe_photon, 
                                     areas = s2_top_area_per_channel, 
                                     mean_pe_photon=self.mean_pe_photon)
                               )
                result[t_+'_2llh'][cur_s2_bool] = np.sum(norm_llh_val, axis=1)

                if self.config['store_per_channel']:
                    store_patterns = np.zeros((s2_pattern.shape[0], self.config['n_top_pmts']) )
                    store_patterns[:, self.pmtbool_top] = s2_pattern
                    result[t_+'_pattern'][cur_s2_bool] = store_patterns#:s2_pattern[cur_s2_bool]

                    store_2LLH_ch = np.zeros((norm_llh_val.shape[0], self.config['n_top_pmts']) )
                    store_2LLH_ch[:, self.pmtbool_top] = norm_llh_val
                    result[t_+'_2llh_per_channel'][cur_s2_bool] = store_2LLH_ch
                    
    @staticmethod
    def _infer_map_format(map_name, known_formats=('pkl', 'json', 'json.gz')):
        for fmt in known_formats:
            if map_name.endswith(fmt):
                return fmt
        raise ValueError(f'Extension of {map_name} not in {known_formats}')


def neg2llh_modpoisson(mu=None, areas=None, mean_pe_photon=1.0):
    """
    Modified poisson distribution with proper normalization for shifted poisson. 
    
    mu - expected number of photons per channel
    areas  - observed areas per channel
    mean_pe_photon - mean of area responce for one photon
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        res = 2.*(mu - 
                  (areas/mean_pe_photon)*np.log(mu) + 
                  loggamma((areas/mean_pe_photon)+1) + 
                  np.log(mean_pe_photon)
                 )
    is_zero = ~(areas>0)    # If area equals or smaller than 0 - assume 0
    res[is_zero] = 2.*mu[is_zero]
    # if zero channel has negative expectation, assume LLH to be 0 there
    # this happens in the normalization factor calculation when mu is received from area
    neg_mu = mu<0.0
    res[is_zero*neg_mu] = 0.0
    return res

# continuous and discrete binomial test
# https://github.com/poliastro/cephes/blob/master/src/bdtr.c

def bdtrc(k, n, p):
    if (k < 0):
        return (1.0)
    if (k == n):
        return (0.0)
    dn = n - k
    if (k == 0):
        if (p < .01):
            dk = -np.expm1(dn * np.log1p(-p))
        else:
            dk = 1.0 - np.exp(dn * np.log(1.0 - p))
    else:
        dk = k + 1
        dk = betainc(dk, dn, p)
    return dk

def bdtr(k, n, p):
    if (k < 0):
        return np.nan
    if (k == n):
        return (1.0)
    dn = n - k
    if (k == 0):
        dk = np.exp(dn * np.log(1.0 - p))
    else:
        dk = k + 1
        dk = betainc(dn, dk, 1.0 - p)
    return dk

# continuous binomial distribution
def binom_pmf(k, n, p):
    scale_log = gammaln(n + 1) - gammaln(n - k + 1) - gammaln(k + 1)
    ret_log = scale_log + k * np.log(p) + (n - k) * np.log(1 - p)
    return np.exp(ret_log)

def binom_cdf(k, n, p):
    return bdtr(k, n, p)

def binom_sf(k, n, p):
    return bdtrc(k, n, p)

def binom_test(k, n, p):
    """
    The main purpose of this algorithm is to find the value j on the
    other side of the mean that has the same probability as k, and
    integrate the tails outward from k and j. In the case where either
    k or j are zero, only the non-zero tail is integrated.
    """
    d = binom_pmf(k, n, p)
    rerr = 1 + 1e-7
    d = d * rerr
    # define number of interaction for finding the the value j
    # the exceptional case of n<=0, is avoid since n_iter is at least 2
    if n > 0:
        n_iter = int(max(np.round(np.log10(n)) + 1, 2))
    else:
        n_iter = 2

    if k < n * p:
        # if binom_pmf(n, n, p) > d, with d<<1e-3, means that we have
        # to look for j value above n. It is likely that the binomial
        # test for such j is extremely low such that we can stop
        # the algorithm and return 0
        if binom_pmf(n, n, p) > d:
            for n_ in np.arange(n, 2*n, 1):
                if binom_pmf(n_, n, p) < d:
                    j_min, j_max = k, n_
                    do_test = True
                    break
                do_test = False
        else:
            j_min, j_max = k, n
            do_test = True
        def _check_(d, y0, y1):
            return (d>y1) and (d<=y0)
    else:
        if binom_pmf(0, n, p) > d:
            n_iter, j_min, j_max = 0, 0, 0
        else:
            j_min, j_max = 0, k
        do_test = True
        def _check_(d, y0, y1):
            return (d>=y0) and (d<y1)

    # if B(k;n,p) is already 0 or I can't find the j in the other side of the mean
    # the returned binomial test is 0
    if (d==0)|(not do_test):
        pval = 0.0
    else:
        # Here we are actually looking for j
        for i in range(n_iter):
            n_pts = int(j_max - j_min)
            if (i<2) and (n_pts < 50): 
                n_pts = 50
            j_range = np.linspace(j_min, j_max, n_pts, endpoint=True)
            y = binom_pmf(j_range, n, p)
            for i in range(len(j_range) - 1):
                if _check_(d, y[i], y[i + 1]):
                    j_min, j_max = j_range[i], j_range[i + 1]
                    break
        j = max(min((j_min + j_max) / 2, n), 0)

        # One side or two side
        # binomial test
        if k * j == 0:
            pval = binom_sf(max(k, j), n, p)
        else:
            pval = binom_cdf(min(k, j), n, p) + binom_sf(max(k, j), n, p)
        pval = min(1.0, pval)
        
    return pval


def _s1_area_fraction_top_probability(aft_prob, area_tot, area_fraction_top, mode='continuous'):
    '''
    Wrapper that does the S1 AFT probability calculation for you
    '''
    
    area_top = area_tot * area_fraction_top
    
    # Raise a warning in case one of these three condition is verified
    # and return binomial test equal to nan since they are not physical
    # k: size_top, n: size_tot, p: aft_prob
    do_test = True
    if area_tot < area_top:
        # warnings.warn(f'n {area_tot} must be >= k {area_top}')
        binomial_test = np.nan
        do_test = False
    if (aft_prob > 1.0) or (aft_prob < 0.0):
        # warnings.warn(f'p {aft_prob} must be in range [0, 1]')
        binomial_test = np.nan
        do_test = False
    if area_top < 0:
        # warnings.warn(f'k {area_top} must be >= 0')
        binomial_test = np.nan
        do_test = False
        
    if do_test:
        if mode == 'discrete':
            binomial_test = binom_pmf(area_top, area_tot, aft_prob)
        else:
            binomial_test = binom_test(area_top, area_tot, aft_prob)
        
    return binomial_test



import strax
import numpy as np
import numba
import straxen

from straxen.get_corrections import is_cmt_option

export, __all__ = strax.exporter()

MV_PREAMBLE = 'Muno-Veto Plugin: Same as the corresponding nVETO-PLugin.\n'


@export
@strax.takes_config(
    strax.Option(
        'save_outside_hits_nv',
        default=(3, 15), track=True,
        help='Save (left, right) samples besides hits; cut the rest'),
    strax.Option(
        'baseline_samples_nv',
        default=('baseline_samples_nv', 'ONLINE', True), track=True,
        help='Number of samples to use at the start of the pulse to determine '
             'the baseline'),
    strax.Option(
        'hit_min_amplitude_nv',
        default=('hit_thresholds_nv', 'ONLINE', True), track=True,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_nveto_pmts, or a number, '
             'or a string like "pmt_commissioning_initial" which means calling '
             'hitfinder_thresholds.py, '
             'or a tuple like (correction=str, version=str, nT=boolean), '
             'which means we are using cmt.'),
    strax.Option(
        'min_samples_alt_baseline_nv',
        default=None, track=True,
        help='Min. length of pulse before alternative baselineing via '
             'pulse median is applied.'),
)
class nVETOPulseProcessing(strax.Plugin):
    """
    nVETO equivalent of pulse processing. The following steps are
    applied:

        1. Flip, baseline and integrate waveforms.
        2. Find hits and apply ZLE
        3. Remove empty fragments.
    """
    __version__ = '0.0.8'

    parallel = 'process'
    rechunk_on_save = False
    compressor = 'zstd'
    save_when = strax.SaveWhen.TARGET

    depends_on = 'raw_records_coin_nv'
    provides = 'records_nv'
    data_kind = 'records_nv'
    ends_with = '_nv'

    def setup(self):
        if isinstance(self.config['baseline_samples_nv'], int):
            self.baseline_samples = self.config['baseline_samples_nv']
        else:
            self.baseline_samples = straxen.get_correction_from_cmt(
                self.run_id, self.config['baseline_samples_nv'])
        
        # Check config of `hit_min_amplitude_nv` and define hit thresholds
        # if cmt config
        if is_cmt_option(self.config['hit_min_amplitude_nv']):
            self.hit_thresholds = straxen.get_correction_from_cmt(self.run_id,
                self.config['hit_min_amplitude_nv'])
        # if hitfinder_thresholds config
        elif isinstance(self.config['hit_min_amplitude_nv'], str):
            self.hit_thresholds = straxen.hit_min_amplitude(
                self.config['hit_min_amplitude_nv'])
        else: # int or array
            self.hit_thresholds = self.config['hit_min_amplitude_nv']

    def infer_dtype(self):
        record_length = strax.record_length_from_dtype(
            self.deps['raw_records_coin_nv'].dtype_for('raw_records_coin_nv'))
        dtype = strax.record_dtype(record_length)
        return dtype

    def compute(self, raw_records_coin_nv):
        # Do not trust in DAQ + strax.baseline to leave the
        # out-of-bounds samples to zero.
        r = strax.raw_to_records(raw_records_coin_nv)
        del raw_records_coin_nv

        r = strax.sort_by_time(r)
        strax.zero_out_of_bounds(r)
        strax.baseline(r,
                       baseline_samples=self.baseline_samples,
                       flip=True)

        if self.config['min_samples_alt_baseline_nv']:
            m = r['pulse_length'] > self.config['min_samples_alt_baseline_nv']
            if np.any(m):
                # Correcting baseline after PMT saturated signals
                r[m] = median_baseline(r[m])

        strax.integrate(r)

        strax.zero_out_of_bounds(r)

        hits = strax.find_hits(r, min_amplitude=self.hit_thresholds)

        le, re = self.config['save_outside_hits_nv']
        r = strax.cut_outside_hits(r, hits, left_extension=le, right_extension=re)
        strax.zero_out_of_bounds(r)

        return r


def median_baseline(records):
    """
    Function which computes the baseline according the pulse's median.

    :param records: Records
    """
    # Count number of pulses
    npulses = np.sum(records['record_i'] == 0)
    fail_counter = 0

    if npulses == 1:
        # This case is simple
        records = _correct_baseline(records)
    else:
        # Now the more complicated case in which we have multiple pulses
        # First we have to group our record fragments into their
        # pulses. Hence get record links and group indicies:
        _, nextr = strax.record_links(records)
        pulse_i = []
        # Loop over the begining of every pulse and get all next indicies.
        for i in np.where(records['record_i'] == 0)[0]:
            inds = [i]
            ind = nextr[i]
            # Always look for next index as long there are some
            while ind != -1:
                inds += [ind]
                ind = nextr[ind]
                fail_counter += 1
                assert fail_counter < 5000, 'Stuck in while-loop pulse is longer than 5000 fragments?!?'

            pulse_i.append(inds)

        for pi in pulse_i:
            records[pi] = _correct_baseline(records[pi])
    return records


@numba.njit
def _correct_baseline(records):
    wf = np.zeros(records[0]['pulse_length'], dtype=np.int16)
    for r in records:
        # np.median(records['data']) does not work for numbafied functions
        # Hence we have to get the entire waveforms first
        if r['record_i'] == 0:
            t0 = r['time']
        i = (r['time'] - t0) // r['dt']
        wf[i:i + r['length']] = r['data'][:r['length']]

    bl = np.median(wf)
    for r in records:
        r['data'][:r['length']] = r['data'][:r['length']] - bl
        r['baseline'] -= bl
    return records


@export
@strax.takes_config(
    strax.Option(
        'save_outside_hits_mv',
        default=(2, 5), track=True,
        child_option=True, parent_option_name='save_outside_hits_nv',
        help='Save (left, right) samples besides hits; cut the rest'),
    strax.Option(
        'baseline_samples_mv',
        default=100, track=True,
        child_option=True, parent_option_name='baseline_samples_nv',
        help='Number of samples to use at the start of the pulse to determine '
             'the baseline'),
    strax.Option(
        'hit_min_amplitude_mv',
        default=('hit_thresholds_mv', 'ONLINE', True), track=True,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_mveto_pmts, or a number, '
             'or a string like "pmt_commissioning_initial" which means calling '
             'hitfinder_thresholds.py, '
             'or a tuple like (correction=str, version=str, nT=boolean),'
             'which means we are using cmt.'),
    strax.Option(
        'check_raw_record_overlaps',
        default=True, track=False,
        help='Crash if any of the pulses in raw_records overlap with others '
             'in the same channel'),
)
class muVETOPulseProcessing(nVETOPulseProcessing):
    __doc__ = MV_PREAMBLE + nVETOPulseProcessing.__doc__
    __version__ = '0.0.1'
    depends_on = 'raw_records_mv'
    provides = 'records_mv'
    data_kind = 'records_mv'
    child_plugin = True

    def setup(self):
        self.baseline_samples = self.config['baseline_samples_mv']

        # Check config of `hit_min_amplitude_mv` and define hit thresholds
        # if cmt config
        if is_cmt_option(self.config['hit_min_amplitude_mv']):
            self.hit_thresholds = straxen.get_correction_from_cmt(self.run_id,
                self.config['hit_min_amplitude_mv'])
        # if hitfinder_thresholds config
        elif isinstance(self.config['hit_min_amplitude_mv'], str):
            self.hit_thresholds = straxen.hit_min_amplitude(
                self.config['hit_min_amplitude_mv'])
        else: # int or array
            self.hit_thresholds = self.config['hit_min_amplitude_mv']

    def infer_dtype(self):
        record_length = strax.record_length_from_dtype(
            self.deps['raw_records_mv'].dtype_for('raw_records_mv'))
        dtype = strax.record_dtype(record_length)
        return dtype

    def compute(self, raw_records_mv):
        if self.config['check_raw_record_overlaps']:
            straxen.check_overlaps(raw_records_mv, n_channels=3000)
        return super().compute(raw_records_mv)



import strax
import straxen
import numpy as np
import numba

export, __all__ = strax.exporter()
MV_PREAMBLE = 'Muno-Veto Plugin: Same as the corresponding nVETO-Plugin.\n'


@export
@strax.takes_config(
    strax.Option(
        'min_veto_area_nv', default=5, type=float, track=True,
        help='Minimal area required in pe to trigger veto.'),
    strax.Option(
        'min_veto_hits_nv', default=10, type=int, track=True,
        help='Minimal number of hitlets in n/mveto_event to trigger a veto.'),
    strax.Option(
        'min_veto_channel_nv', default=0, type=int, track=True,
        help='Minimal number PMT channel contributing to the n/mveto_event.'),
    strax.Option(
        'veto_left_extension_nv', default=500_000, type=int, track=True,
        help='Veto time in ns left t the start of a vetoing event.'),
    strax.Option(
        'veto_right_extension_nv', default=0, type=int, track=True,
        help='Veto time in ns right to the end of a vetoing event.'),
)
class nVETOVetoRegions(strax.OverlapWindowPlugin):
    """
    Plugin which defines the time intervals in which peaks should be
    tagged as vetoed. An event must surpass all three criteria to trigger
    a veto.
    """
    __version__ = '0.0.1'

    depends_on = 'events_nv'
    provides = 'veto_regions_nv'
    data_kind = 'veto_regions_nv'
    save_when = strax.SaveWhen.NEVER

    dtype = strax.time_fields
    ends_with = '_nv'

    def get_window_size(self):
        return 10 * (self.config['veto_left_extension_nv'] + self.config['veto_right_extension_nv'])

    def compute(self, events_nv, start, end):
        vetos = create_veto_intervals(events_nv,
                                      self.config['min_veto_area_nv'],
                                      self.config['min_veto_hits_nv'],
                                      self.config['min_veto_channel_nv'],
                                      self.config['veto_left_extension_nv'],
                                      self.config['veto_right_extension_nv'])

        # Now we have to do clip all times and end endtimes in case
        # they go beyond a chunk boundary:
        vetos['time'] = np.clip(vetos['time'], start, end)
        vetos['endtime'] = np.clip(vetos['endtime'], start, end)

        return vetos


def create_veto_intervals(events,
                          min_area,
                          min_hits,
                          min_contributing_channels,
                          left_extension,
                          right_extension, ):
    """
    Function which creates veto regions.

    :param events: nveto events
    :param min_area: min area in pe required to create a veto region.
    :param min_hits: same but with hitlets
    :param min_contributing_channels: Minimal number of contributing
        channels.
    :param left_extension: Left extension of the event to define veto
        region in ns.
    :param right_extension: Same but right hand side after.

    :returns: numpy.structured.array containing the veto regions.
    """
    res = np.zeros(len(events),
                   dtype=strax.time_fields)

    res = _create_veto_intervals(events,
                                 min_area,
                                 min_hits,
                                 min_contributing_channels,
                                 left_extension,
                                 right_extension,
                                 res, )
    res = straxen.merge_intervals(res)

    return res


@numba.njit(cache=True)
def _create_veto_intervals(events,
                           min_area,
                           min_hits,
                           min_contributing_channels,
                           left_extension,
                           right_extension,
                           res, ):
    offset = 0

    for ev in events:
        satisfies_veto_trigger = (ev['area'] >= min_area
                                  and ev['n_hits'] >= min_hits
                                  and ev['n_contributing_pmt'] >= min_contributing_channels)
        if not satisfies_veto_trigger:
            continue

        res[offset]['time'] = ev['time'] - left_extension
        res[offset]['endtime'] = ev['endtime'] + right_extension
        offset += 1
    return res[:offset]


@strax.takes_config(
    strax.Option(
        'min_veto_area_mv', default=10, type=float, track=True,
        child_option=True, parent_option_name='min_veto_area_nv',
        help='Minimal area required in pe to trigger veto.'),
    strax.Option(
        'min_veto_hits_mv', default=0, type=int, track=True,
        child_option=True, parent_option_name='min_veto_hits_nv',
        help='Minimal number of hitlets in event to trigger veto.'),
    strax.Option(
        'min_veto_channel_mv', default=5, type=int, track=True,
        child_option=True, parent_option_name='min_veto_channel_nv',
        help='Minimal number PMT channel contributing to the event.'),
    strax.Option(
        'veto_left_extension_mv', default=0, type=int, track=True,
        child_option=True, parent_option_name='veto_left_extension_nv',
        help='Veto time in ns left t the start of a vetoing event.'),
    strax.Option(
        'veto_right_extension_mv', default=1_000_000, type=int, track=True,
        child_option=True, parent_option_name='veto_right_extension_nv',
        help='Veto time in ns right to the end of a vetoing event.'),
)
class muVETOVetoRegions(nVETOVetoRegions):
    __doc__ = MV_PREAMBLE + nVETOVetoRegions.__doc__
    __version__ = '0.0.1'

    depends_on = 'events_mv'
    provides = 'veto_regions_mv'
    data_kind = 'veto_regions_mv'
    save_when = strax.SaveWhen.NEVER

    dtype = strax.time_fields
    ends_with = '_mv'
    child_plugin = True

    def get_window_size(self):
        return 10 * (self.config['veto_left_extension_mv'] + self.config['veto_left_extension_mv'])

    def compute(self, events_mv, start, end):
        return super().compute(events_mv, start, end)



"""
XENON1T cuts

How to apply:
 - First register the plugins:
 st.register_all(straxen.plugins.x1t_cuts)
 - Load events and the cuts you want to apply:
 events = st.get_array(run_id,
                       targets=('event_info', 'cut_s2_width', 'cut_s2_threshold'))
 - Apply the selection of events that pass the cut like:
 selected_events = events[events['cut_s2_threshold'] == True]
"""

import numpy as np
from scipy.stats import chi2
import strax

from straxen import units

export, __all__ = strax.exporter()


class S2Width(strax.CutPlugin):
    """S2 Width cut based on diffusion model
    The S2 width cut compares the S2 width to what we could expect based on its depth in the detector. The inputs to
    this are the drift velocity and the diffusion constant. The allowed variation in S2 width is greater at low
    energy (since it is fluctuating statistically) Ref: (arXiv:1102.2865)
    It should be applicable to data regardless of if it ER or NR;
    above cS2 = 1e5 pe ERs the acceptance will go down due to track length effects.
    around S2 = 1e5 pe there are beta-gamma merged peaks from Pb214 that extends the S2 width
    Tune the diffusion model parameters based on fax data according to note:
    https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenon1t:sim:notes:tzhu:width_cut_tuning#toy_fax_simulation
    Contact: Tianyu <tz2263@columbia.edu>, Yuehuan <weiyh@physik.uzh.ch>, Jelle <jaalbers@nikhef.nl>
    ported from lax.sciencerun1.py"""
    depends_on = ('event_info',)
    provides = 'cut_s2_width'
    cut_name = 'cut_s2_width'
    cut_description = 'S2 Width cut'

    __version__ = 1.3

    diffusion_constant = 29.35 * ((units.cm) ** 2) / units.s
    v_drift = 1.335 * (units.um) / units.ns
    scg = 21.3  # s2_secondary_sc_gain in pax config
    scw = 229.58  # s2_secondary_sc_width median
    SigmaToR50 = 1.349
    DriftTimeFromGate = 1.6 * units.us

    def s2_width_model(self, drift_time):
        """Diffusion model
        """
        return np.sqrt(2 * self.diffusion_constant * (drift_time - self.DriftTimeFromGate) / self.v_drift ** 2)

    def nElectron(self, events):
        return np.clip(events['s2_area'], 0, 5000) / self.scg

    def normWidth(self, events):
        return (np.square(events['s2_range_50p_area'] / self.SigmaToR50) - np.square(self.scw)) / \
               np.square(self.s2_width_model(events['drift_time']))

    def logpdf(self, events):
        return chi2.logpdf(self.normWidth(events) * (self.nElectron(events) - 1), self.nElectron(events))

    def cut_by(self, events):
        return np.all([self.logpdf(events) > -14], axis=0)


class S1SingleScatter(strax.CutPlugin):
    """Requires only one valid interaction between the largest S2, and any S1 recorded before it.
    The S1 cut checks that any possible secondary S1s recorded in a waveform, could not have also
    produced a valid interaction with the primary S2. To check whether an interaction between the
    second largest S1 and the largest S2 is valid, we use the S2Width cut. If the event would pass
    the S2Width cut, a valid second interaction exists, and we may have mis-identified which S1 to
    pair with the primary S2. Therefore we cut this event. If it fails the S2Width cut the event is
    not removed.
    Current version is developed on calibration data (pax v6.8.0). It is described in this note:
    https://xecluster.lngs.infn.it/dokuwiki/doku.php?id=xenon:xenon1t:jacques:s1_single_scatter_cut_sr1
    It should be applicable to data regardless whether it is ER or NR.
    Contact: Jacques Pienaar, <jpienaar@uchicago.edu>
    ported from lax.sciencerun1.py
    """
    depends_on = 'event_info'
    provides = 'cut_s1_single_scatter'
    cut_name = 'cut_s1_single_scatter'
    cut_description = 'S1 Single Scatter cut'
    s2width = S2Width

    __version__ = 1.3

    def cut_by(self, events):
        mask = events['alt_s1_interaction_drift_time'] > self.s2width.DriftTimeFromGate
        alt_n_electron = np.clip(events[mask]['s2_area'], 0, 5000) / self.s2width.scg

        # Alternate S1 relative width
        alt_rel_width = np.square(events[mask]['s2_range_50p_area'] / self.s2width.SigmaToR50) - np.square(
            self.s2width.scw)
        alt_rel_width /= np.square(self.s2width.s2_width_model(self.s2width,
                                                               events[mask]['alt_s1_interaction_drift_time']))

        alt_interaction_passes = chi2.logpdf(alt_rel_width * (alt_n_electron - 1), alt_n_electron) > - 20

        return np.all([True ^ alt_interaction_passes], axis=0)


class S2SingleScatter(strax.CutPlugin):
    """Check that largest other S2 area is smaller than some bound.
    The single scatter is to cut an event if its largest_other_s2 is too large.
    As the largest_other_s2 takes a greater value when they originated from some real scatters
    in comparison, those from photo-ionization in single scatter cases would be smaller.
    https://xecluster.lngs.infn.it/dokuwiki/doku.php?id=xenon:xenon1t:analysis:firstresults:cut:s2single
    Contact: Tianyu Zhu <tz2263@columbia.edu>
    ported from lax.sciencerun1.py
    """
    depends_on = 'event_info'
    provides = 'cut_s2_single_scatter'
    cut_name = 'cut_s2_single_scatter'
    cut_description = 'S2 Single Scatter cut'
    __version__ = 4.3

    @classmethod
    def other_s2_bound(cls, s2_area):
        rescaled_s2_0 = s2_area * 0.00832 + 72.3
        rescaled_s2_1 = s2_area * 0.03 - 109

        another_term_0 = 1 / (np.exp((s2_area - 23300) * 5.91e-4) + 1)
        another_term_1 = 1 / (np.exp((23300 - s2_area) * 5.91e-4) + 1)

        return rescaled_s2_0 * another_term_0 + rescaled_s2_1 * another_term_1

    def cut_by(self, events):
        largest_other_s2_is_nan = np.isnan(events['alt_s2_area'])
        arr = np.all([largest_other_s2_is_nan | (events['alt_s2_area'] < self.other_s2_bound(events['s2_area']))],
                     axis=0)
        return arr


class S2Threshold(strax.CutPlugin):
    """The S2 energy at which the trigger is perfectly efficient.
    See: https://xecluster.lngs.infn.it/dokuwiki/doku.php?id=xenon:xenon1t:analysis:firstresults:daqtriggerpaxefficiency
    Contact: Jelle Aalbers <aalbers@nikhef.nl>
    ported from lax.sciencerun1.py
    """
    depends_on = 'event_info'
    provides = 'cut_s2_threshold'
    cut_name = 'cut_s2_threshold'
    cut_description = 's2 must be larger then 200 PE'
    __version__ = 1.4

    def cut_by(self, events):
        return np.all([events['s2_area'] > 200], axis=0)


class S2AreaFractionTop(strax.CutPlugin):
    """Cuts events with an unusual fraction of S2 on top array.
    Primarily cuts gas events with a particularly large S2 AFT, also targets some
    strange / junk / other events with a low AFT.
    This cut has been checked on S2 ranges between 0 and 50 000 pe.
    Described in the note at: xenon:xenon1t:analysis:firstresults:s2_aft_cut_summary
    Contact: Adam Brown <abrown@physik.uzh.ch>,
    ported from lax.sciencerun1.py
    """
    depends_on = ('event_info',)
    provides = 'cut_s2_area_fraction_top'
    cut_name = 'cut_s2_area_fraction_top'
    cut_description = 'Cut on S2 AFT'

    __version__ = 1.3

    def upper_limit_s2_aft(self, s2):
        return 0.6177399420527526 + 3.713166211522462e-08 * s2 + 0.5460484265254656 / np.log(s2)

    def lower_limit_s2_aft(self, s2):
        return 0.6648160611018054 - 2.590402853814859e-07 * s2 - 0.8531029789184852 / np.log(s2)

    def cut_by(self, events):
        arr = np.all([events['s2_area_fraction_top'] < self.upper_limit_s2_aft(events['s2_area']),
                      events['s2_area_fraction_top'] > self.lower_limit_s2_aft(events['s2_area'])], axis=0)
        return arr


class FiducialCylinder1T(strax.CutPlugin):
    """Implementation of fiducial volume cylinder 1T,
    ported from lax.sciencerun0.py"""
    depends_on = ('event_positions',)
    provides = 'cut_fiducial_cylinder_1t'
    cut_name = 'cut_fiducial_cylinder_1t'
    cut_description = 'One tonne fiducial cylinder'

    __version__ = 1.3

    def cut_by(self, events):
        return np.all([(-92.9 < events['z']), (-9 > events['z']),
                       (36.94 > np.sqrt(events['x'] ** 2 + events['y'] ** 2))],
                      axis=0)


class S1MaxPMT(strax.LoopPlugin):
    """Removes events where the largest hit in S1 is too large
    port from lax.sciencerun0.py"""
    depends_on = ('events', 'event_basics', 'peak_basics')
    dtype = [('cut_s1_max_pmt', np.bool, 'S1 max PMT cut')] + strax.time_fields
    provides = 'cut_s1_max_pmt'
    __version__ = 1.2

    def compute_loop(self, event, peaks):
        ret = dict(cut_s1_max_pmt=True)
        ret['time'] = event['time']
        ret['endtime'] = event['endtime']
        if not len(peaks) or np.isnan(event['s1_index']):
            return ret

        peak = peaks[event['s1_index']]
        max_channel = peak['max_pmt_area']
        ret['cut_s1_max_pmt'] = (max_channel < 0.052 * event['s1_area'] + 4.15)
        return ret


class S1LowEnergyRange(strax.CutPlugin):
    """Pass only events with cs1<200"""
    depends_on = ('events', 'corrected_areas')
    provides = 'cut_s1_low_energy_range'
    cut_name = 'cut_s1_low_energy_range'
    cut_description = "Event under 200pe"

    __version__ = 1.3

    def cut_by(self, events):
        return np.all([events['cs1'] < 200], axis=0)


class SR1Cuts(strax.MergeOnlyPlugin):
    depends_on = ['cut_fiducial_cylinder_1t', 'cut_s1_max_pmt', 'cut_s1_low_energy_range']
    save_when = strax.SaveWhen.ALWAYS


class FiducialEvents(strax.Plugin):
    depends_on = ['event_info', 'cut_fiducial_cylinder_1t']
    data_kind = 'fiducial_events'

    def infer_dtype(self):
        dtype = [self.deps[d].dtype_for(d) for d in self.depends_on]
        dtype.sort()
        return strax.merged_dtype(dtype)

    def compute(self, events):
        return events[events['cut_fiducial_cylinder_1t']]



import inspect
import textwrap

import pandas as pd
import strax
from strax.context import select_docs
import straxen


export, __all__ = strax.exporter()

ma_doc = """
This is a straxen mini-analysis.
The method takes run_id as its only positional argument,
and additional arguments through keywords only.

The function requires the data types: {requires}.
Unless you specify this through data_kind = array keyword arguments,
this data will be loaded automatically. 

The function takes the same selection arguments as context.get_array:
""" + select_docs

_hv_bokeh_initialized = False


@export
def mini_analysis(requires=tuple(),
                  hv_bokeh=False,
                  warn_beyond_sec=None,
                  default_time_selection='touching'):
    def decorator(f):
        parameters = inspect.signature(f).parameters

        def wrapped_f(context: strax.Context, run_id: str, **kwargs):
            # Validate arguments
            known_kwargs = (
                'time_range seconds_range time_within time_selection '
                'ignore_time_warning '
                'selection_str t_reference to_pe config').split()
            for k in kwargs:
                if k not in known_kwargs and k not in parameters:
                    # Python itself also raises TypeError for invalid kwargs
                    raise TypeError(f"Unknown argument {k} for {f.__name__}")

            if 'config' in kwargs:
                context = context.new_context(config=kwargs['config'])
            if 'config' in parameters:
                kwargs['config'] = context.config

            # Say magic words to enable holoviews
            if hv_bokeh:
                global _hv_bokeh_initialized
                if not _hv_bokeh_initialized:
                    import holoviews
                    holoviews.extension('bokeh')
                    _hv_bokeh_initialized = True

            # TODO: This is a placeholder until the corrections system
            # is more fully developed
            if 'to_pe' in parameters and 'to_pe' not in kwargs:
                kwargs['to_pe'] = straxen.get_correction_from_cmt(
                    run_id,
                    context.config['gain_model'])

            # Prepare selection arguments
            kwargs['time_range'] = context.to_absolute_time_range(
                run_id,
                targets=requires,
                **{k: kwargs.get(k)
                   for k in ('time_range seconds_range time_within'.split())})
            kwargs.setdefault('time_selection', default_time_selection)
            kwargs.setdefault('selection_str', None)

            kwargs['t_reference'], _ = context.estimate_run_start_and_end(
                run_id, requires)

            if warn_beyond_sec is not None and not kwargs.get('ignore_time_warning'):
                tr = kwargs['time_range']
                if tr is None:
                    sec_requested = float('inf')
                else:
                    sec_requested = (tr[1] - tr[0]) / int(1e9)
                if sec_requested > warn_beyond_sec:
                    tr_str = "the entire run" if tr is None else f"{sec_requested} seconds"
                    raise ValueError(
                        f"The author of this mini analysis recommends "
                        f"not requesting more than {warn_beyond_sec} seconds. "
                        f"You are requesting {tr_str}. If you wish to proceed, "
                        "pass ignore_time_warning = True.")

            # Load required data, if any
            if len(requires):
                deps_by_kind = strax.group_by_kind(
                    requires, context=context)
                for dkind, dtypes in deps_by_kind.items():
                    if dkind in kwargs:
                        # Already have data, just apply cuts
                        kwargs[dkind] = context.apply_selection(
                            kwargs[dkind],
                            selection_str=kwargs['selection_str'],
                            time_range=kwargs['time_range'],
                            time_selection=kwargs['time_selection'])
                    else:
                        kwargs[dkind] = context.get_array(
                            run_id,
                            dtypes,
                            selection_str=kwargs['selection_str'],
                            time_range=kwargs['time_range'],
                            time_selection=kwargs['time_selection'],
                            # Arguments for new context, if needed
                            config=kwargs.get('config'),
                            register=kwargs.get('register'),
                            storage=kwargs.get('storage', tuple()),
                            progress_bar=False,
                        )

                # If user did not give time kwargs, but the function expects
                # a time_range, try to add one based on the time range of the data
                base_dkind = list(deps_by_kind.keys())[0]
                x = kwargs[base_dkind]
                if len(x) and kwargs.get('time_range') is None:
                    x0 = x.iloc[0] if isinstance(x, pd.DataFrame) else x[0]
                    try:
                        kwargs.setdefault('time_range', (x0['time'],
                                                         strax.endtime(x).max()))

                    except AttributeError:
                        # If x is a holoviews dataset, this will fail.
                        pass

            if 'seconds_range' in parameters:
                if kwargs.get('time_range') is None:
                    scr = None
                else:
                    scr = tuple([(t - kwargs['t_reference']) / int(1e9)
                                 for t in kwargs['time_range']])
                kwargs.setdefault('seconds_range', scr)

            kwargs.setdefault('run_id', run_id)
            kwargs.setdefault('context', context)

            if 'kwargs' in parameters:
                # Likely this will be passed to another mini-analysis
                to_pass = kwargs
                # Do not pass time_range and seconds_range both (unless explicitly requested)
                # strax does not like that
                if 'seconds_range' in to_pass and not 'seconds_range' in parameters:
                    del to_pass['seconds_range']
                if 'time_within' in to_pass and not 'time_within' in parameters:
                    del to_pass['time_within']
            else:
                # Pass only arguments the function wants
                to_pass = {k: v for k, v in kwargs.items()
                           if k in parameters}
            return f(**to_pass)

        wrapped_f.__name__ = f.__name__

        if hasattr(f, '__doc__') and f.__doc__:
            doc_lines = f.__doc__.splitlines()
            wrapped_f.__doc__ = (doc_lines[0]
                               + '\n'
                               + textwrap.dedent(
                        '\n'.join(doc_lines[1:])))
        else:
            wrapped_f.__doc__ = \
                ('Straxen mini-analysis for which someone was too lazy'
                 'to write a proper docstring')

        wrapped_f.__doc__ += ma_doc.format(requires=', '.join(requires))

        strax.Context.add_method(wrapped_f)
        return wrapped_f

    return decorator



import numpy as np
import straxen
import warnings


def test_query_sc_values():
    """
    Unity test for the SCADAInterface. Query a fixed range and check if 
    return is correct.
    """
    if not straxen.utilix_is_configured('scada', 'scdata_url'):
        warnings.warn('Cannot test scada since we have no access to xenon secrets.')
        return

    print('Testing SCADAInterface')
    sc = straxen.SCADAInterface(use_progress_bar=False)

    print('Query single value:')
    # Simple query test:
    # Query 5 s of data:
    start = 1609682275000000000
    end = start + 5 * 10**9
    parameters = {'SomeParameter': 'XE1T.CTPC.Board06.Chan011.VMon'}

    df = sc.get_scada_values(parameters,
                             start=start,
                             end=end,
                             every_nth_value=1,
                             query_type_lab=False,)

    assert df['SomeParameter'][0] // 1 == 1253, 'First values returned is not corrrect.'
    assert np.all(np.isnan(df['SomeParameter'][1:])), 'Subsequent values are not correct.'

    # Test ffill option:
    print('Testing forwardfill option:')
    parameters = {'SomeParameter': 'XE1T.CRY_FCV104FMON.PI'}
    df = sc.get_scada_values(parameters,
                             start=start,
                             end=end,
                             fill_gaps='forwardfill',
                             every_nth_value=1,
                             query_type_lab=False,)
    assert np.all(np.isclose(df[:4], 2.079859)), 'First four values deviate from queried values.'
    assert np.all(np.isclose(df[4:], 2.117820)), 'Last two values deviate from queried values.'

    print('Testing downsampling and averaging option:')
    parameters = {'SomeParameter': 'XE1T.CRY_TE101_TCRYOBOTT_AI.PI'}
    df_all = sc.get_scada_values(parameters,
                                 start=start,
                                 end=end,
                                 fill_gaps='forwardfill',
                                 every_nth_value=1,
                                 query_type_lab=False,)

    df = sc.get_scada_values(parameters,
                             start=start,
                             end=end,
                             fill_gaps='forwardfill',
                             down_sampling=True,
                             every_nth_value=2,
                             query_type_lab=False,)

    assert np.all(df_all[::2] == df), 'Downsampling did not return the correct values.'

    df = sc.get_scada_values(parameters,
                             start=start,
                             end=end,
                             fill_gaps='forwardfill',
                             every_nth_value=2,
                             query_type_lab=False,)

    # Compare average for each second value:
    for ind, i in enumerate([0, 2, 4]):
        assert np.isclose(np.mean(df_all[i:i + 2]), df['SomeParameter'][ind]), 'Averaging is incorrect.'

    # Testing lab query type:
    df = sc.get_scada_values(parameters,
                             start=start,
                             end=end,
                             query_type_lab=True,)

    assert np.all(df['SomeParameter'] // 1 == -96), 'Not all values are correct for query type lab.'



from straxen import _is_on_pytest


def test_testing_suite():
    """
    Make sure that we are always on a pytest when this is called. E.g.
        pytest tests/test_testing_suite.py

    A counter example would e.g. be:
        python tests/test_testing_suite.py
    """
    assert _is_on_pytest()


if __name__ == '__main__':
    test_testing_suite()



"""Test several functions distibuted over common.py, misc.py, scada.py"""
import straxen
import pandas
import os
import tempfile
from .test_basics import test_run_id_1T
import numpy as np
import strax
from matplotlib.pyplot import clf as plt_clf


def test_pmt_pos_1t():
    """
    Test if we can get the 1T PMT positions
    """
    pandas.DataFrame(straxen.pmt_positions(True))


def test_pmt_pos_nt():
    """
    Test if we can get the nT PMT positions
    """
    pandas.DataFrame(straxen.pmt_positions(False))


def test_secret():
    """
    Check something in the sectets. This should not work because we
    don't have any.
    """
    try:
        straxen.get_secret('somethingnonexistent')
    except ValueError:
        # Good we got some message we cannot load something that does
        # not exist,
        pass


# If one of the test below fail, perhaps these values need to be updated.
# They were added on 27/11/2020 and may be outdated by now
EXPECTED_OUTCOMES_TEST_SEVERAL = {
    'n_peaks': 128,
    'n_s1': 6,
    'run_live_time': 0.17933107,
    'n_events': 2,
}


def test_several():
    """
    Test several other functions in straxen. Is kind of messy but saves
    time as we won't load data many times
    :return:
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            print("Temporary directory is ", temp_dir)
            os.chdir(temp_dir)

            print("Downloading test data (if needed)")
            st = straxen.contexts.demo()
            st.make(test_run_id_1T, 'records')
            # Ignore strax-internal warnings
            st.set_context_config({'free_options': tuple(st.config.keys())})
            st.make(test_run_id_1T, 'records')

            print("Get peaks")
            p = st.get_array(test_run_id_1T, 'peaks')

            # Do checks on there number of peaks
            assertion_statement = ("Got /more peaks than expected, perhaps "
                                   "the test is outdated or clustering has "
                                   "really changed")
            assert np.abs(len(p) -
                          EXPECTED_OUTCOMES_TEST_SEVERAL['n_peaks']) < 5, assertion_statement

            events = st.get_array(test_run_id_1T, 'event_info')
            print('plot wf')
            peak_i = 0
            st.plot_waveform(test_run_id_1T, time_range=(p[peak_i]['time'], strax.endtime(p[peak_i])))
            plt_clf()

            print('plot hit pattern')
            peak_i = 1
            st.plot_hit_pattern(test_run_id_1T, time_range=(p[peak_i]['time'], strax.endtime(p[peak_i])), xenon1t=True)
            plt_clf()

            print('plot (raw)records matrix')
            peak_i = 2
            assert st.is_stored(test_run_id_1T, 'records'), "no records"
            assert st.is_stored(test_run_id_1T, 'raw_records'), "no raw records"
            st.plot_records_matrix(test_run_id_1T, time_range=(p[peak_i]['time'],
                                                               strax.endtime(p[peak_i])))

            st.raw_records_matrix(test_run_id_1T, time_range=(p[peak_i]['time'],
                                                           strax.endtime(p[peak_i])))
            st.plot_waveform(test_run_id_1T,
                             time_range=(p[peak_i]['time'],
                                         strax.endtime(p[peak_i])),
                             deep=True)
            plt_clf()

            straxen.analyses.event_display.plot_single_event(st,
                                                             test_run_id_1T,
                                                             events,
                                                             xenon1t=True,
                                                             event_number=0,
                                                             records_matrix='raw')
            st.event_display_simple(test_run_id_1T,
                                    time_range=(events[0]['time'],
                                                events[0]['endtime']),
                                    xenon1t=True)
            plt_clf()

            st.event_display_interactive(test_run_id_1T, time_range=(events[0]['time'],
                                                                  events[0]['endtime']),
                                         xenon1t=True)
            plt_clf()

            print('plot aft')
            st.plot_peaks_aft_histogram(test_run_id_1T)
            plt_clf()

            print('plot event scatter')
            st.event_scatter(test_run_id_1T)
            plt_clf()

            print('plot event scatter')
            st.plot_energy_spectrum(test_run_id_1T)
            plt_clf()

            print('plot peak clsassification')
            st.plot_peak_classification(test_run_id_1T)
            plt_clf()

            print("plot holoviews")
            peak_i = 3
            st.waveform_display(test_run_id_1T,
                                time_range=(p[peak_i]['time'],
                                            strax.endtime(p[peak_i])))
            st.hvdisp_plot_pmt_pattern(test_run_id_1T,
                                time_range=(p[peak_i]['time'],
                                            strax.endtime(p[peak_i])))
            st.hvdisp_plot_peak_waveforms(test_run_id_1T,
                                time_range=(p[peak_i]['time'],
                                            strax.endtime(p[peak_i])))


            print('Plot single pulse:')
            st.plot_pulses_tpc(test_run_id_1T, max_plots=2,  plot_hits=True, ignore_time_warning=True)

            print("Check live-time")
            live_time = straxen.get_livetime_sec(st, test_run_id_1T, things=p)
            assertion_statement = "Live-time calculation is wrong"
            assert live_time == EXPECTED_OUTCOMES_TEST_SEVERAL['run_live_time'], assertion_statement

            print('Check the peak_basics')
            df = st.get_df(test_run_id_1T, 'peak_basics')
            assertion_statement = ("Got less/more S1s than expected, perhaps "
                                   "the test is outdated or classification "
                                   "has really changed.")
            assert np.abs(np.sum(df['type'] == 1) -
                          EXPECTED_OUTCOMES_TEST_SEVERAL['n_s1']) < 2, assertion_statement
            df = df[:10]

            print("Check that we can write nice wiki dfs")
            straxen.dataframe_to_wiki(df)

            print("Abuse the peaks to show that _average_scada works")
            p = p[:10]
            p_t, p_a = straxen.scada._average_scada(
                p['time']/1e9,
                p['time'],
                1)
            assert len(p_a) == len(p), 'Scada deleted some of my 10 peaks!'

            print('Check the number of events')
            events = st.get_array(test_run_id_1T, 'event_info_double')
            assertion_statement = ("Got less/ore events than expected, "
                                   "perhaps the test is outdated or something "
                                   "changed in the processing.")
            assert len(events) == EXPECTED_OUTCOMES_TEST_SEVERAL['n_events'], assertion_statement

            print("Plot bokkeh:")
            fig = st.event_display_interactive(test_run_id_1T,
                                               time_range=(events[0]['time'],
                                                           events[0]['endtime']),
                                               xenon1t=True,
                                               plot_record_matrix=True,
                                               )
            fig.save('test_display.html')

            # Test data selector:
            from straxen.analyses.bokeh_waveform_plot import DataSelectionHist
            ds = DataSelectionHist('ds')
            fig = ds.histogram2d(p,
                               p['area'],
                               p['area'],
                               bins=50,
                               hist_range=((0, 200), (0, 2000)),
                               log_color_scale=True,
                               clim=(10, None),
                               undeflow_color='white')

            import bokeh.plotting as bklt
            bklt.save(fig, 'test_data_selector.html')

        # On windows, you cannot delete the current process'
        # working directory, so we have to chdir out first.
        finally:
            os.chdir('..')


def test_plots():
    """Make some plots"""
    c = np.ones(straxen.n_tpc_pmts)
    straxen.plot_pmts(c)
    straxen.plot_pmts(c, log_scale=True)


def test_print_version():
    straxen.print_versions(['strax', 'something_that_does_not_exist'])


def test_nt_minianalyses():
    """Number of tests to be run on nT like configs"""
    if not straxen.utilix_is_configured():
        return
    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            print("Temporary directory is ", temp_dir)
            os.chdir(temp_dir)
            from .test_plugins import DummyRawRecords, testing_config_nT, test_run_id_nT
            st = straxen.contexts.xenonnt_online(use_rucio=False)
            rundb = st.storage[0]
            rundb.readonly = True
            st.storage = [rundb, strax.DataDirectory(temp_dir)]

            # We want to test the FDC map that only works with CMT
            test_conf = testing_config_nT.copy()
            del test_conf['fdc_map']

            st.set_config(test_conf)
            st.set_context_config(dict(forbid_creation_of=()))
            st.register(DummyRawRecords)

            rr = st.get_array(test_run_id_nT, 'raw_records')
            st.make(test_run_id_nT, 'records')
            st.make(test_run_id_nT, 'peak_basics')

            st.daq_plot(test_run_id_nT,
                        time_range=(rr['time'][0], strax.endtime(rr)[-1]),
                        vmin=0.1,
                        vmax=1,
                        )

            st.plot_records_matrix(test_run_id_nT,
                                   time_range=(rr['time'][0],
                                               strax.endtime(rr)[-1]),
                                   vmin=0.1,
                                   vmax=1,
                                   group_by='ADC ID',
                                   )
            plt_clf()

            st.make(test_run_id_nT, 'event_info')
            st.load_corrected_positions(test_run_id_nT,
                                        time_range=(rr['time'][0],
                                                    strax.endtime(rr)[-1]),

                                        )
            # This would be nice to add but with empty events it does not work
            # st.event_display(test_run_id_nT,
            #                  time_range=(rr['time'][0],
            #                              strax.endtime(rr)[-1]),
            #                  )
        # On windows, you cannot delete the current process'git p
        # working directory, so we have to chdir out first.
        finally:
            os.chdir('..')



import setuptools


def open_requirements(path):
    with open(path) as f:
        requires = [
            r.split('/')[-1] if r.startswith('git+') else r
            for r in f.read().splitlines()]
    return requires


# Get requirements from requirements.txt, stripping the version tags
requires = open_requirements('requirements.txt')
tests_requires = open_requirements('extra_requirements/requirements-tests.txt')
doc_requirements = open_requirements('extra_requirements/requirements-docs.txt')

with open('README.md') as file:
    readme = file.read()

with open('HISTORY.md') as file:
    history = file.read()

setuptools.setup(name='straxen',
                 version='0.19.2',
                 description='Streaming analysis for XENON',
                 author='Straxen contributors, the XENON collaboration',
                 url='https://github.com/XENONnT/straxen',
                 long_description=readme + '\n\n' + history,
                 long_description_content_type="text/markdown",
                 setup_requires=['pytest-runner'],
                 install_requires=requires,
                 tests_require=requires + tests_requires,
                 python_requires=">=3.6",
                 extras_require={
                     'docs': doc_requirements,
                     'microstrax': ['hug'],
                 },
                 scripts=[
                     'bin/bootstrax',
                     'bin/straxer',
                     'bin/fake_daq',
                     'bin/microstrax',
                     'bin/ajax',
                     'bin/refresh_raw_records',
                 ],
                 packages=setuptools.find_packages() + ['extra_requirements'],
                 package_dir={'extra_requirements': 'extra_requirements'},
                 package_data={'extra_requirements': ['requirements-docs.txt',
                                                      'requirements-tests.txt']},
                 classifiers=[
                     'Development Status :: 5 - Production/Stable',
                     'License :: OSI Approved :: BSD License',
                     'Natural Language :: English',
                     'Programming Language :: Python :: 3.6',
                     'Programming Language :: Python :: 3.7',
                     'Programming Language :: Python :: 3.8',
                     'Programming Language :: Python :: 3.9',
                     'Intended Audience :: Science/Research',
                     'Programming Language :: Python :: Implementation :: CPython',
                     'Topic :: Scientific/Engineering :: Physics',
                 ],
                 zip_safe=False)



import tempfile
import os

import numpy as np
import straxen

test_run_id_1T = '180423_1021'


def test_straxen():
    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            print("Temporary directory is ", temp_dir)
            os.chdir(temp_dir)

            print("Downloading test data (if needed)")
            st = straxen.contexts.demo()
            # Ignore strax-internal warnings
            st.set_context_config({'free_options': tuple(st.config.keys())})

            run_df = st.select_runs(available='raw_records')
            print(run_df)
            run_id = run_df.iloc[0]['name']
            assert run_id == test_run_id_1T

            print("Test processing")
            df = st.get_df(run_id, 'event_info')

            assert len(df) > 0
            assert 'cs1' in df.columns
            assert df['cs1'].sum() > 0
            assert not np.all(np.isnan(df['x'].values))

            print('Test common.get_livetime_sec')
            events = st.get_array(run_id, 'peaks')
            straxen.get_livetime_sec(st, test_run_id_1T, things=events)
            # TODO: find a way to break up the tests
            # surely pytest has common startup/cleanup?

            print("Test mini analysis")
            @straxen.mini_analysis(requires=('raw_records',))
            def count_rr(raw_records):
                return len(raw_records)

            n = st.count_rr(test_run_id_1T)
            assert n > 100

        # On windows, you cannot delete the current process'
        # working directory, so we have to chdir out first.
        finally:
            os.chdir('..')
